# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Ignore false-positives for redefined `product` functions:
# pylint: disable=function-redefined

import itertools

import numpy as np

from netket.utils import HashableArray, struct
from netket.utils.types import Array, DType, Shape
from netket.utils.dispatch import dispatch

from netket.utils import warn_deprecation, deprecated, deprecated_new_name

from ._group import FiniteGroup
from ._semigroup import Element


class Permutation(Element):
    def __init__(
        self,
        permutation: Array | None = None,
        *,  # change one line somewhere
        name: str | None = None,
        permutation_array: Array | None = None,
        inverse_permutation_array: Array | None = None,
    ):
        r"""
        Creates a `Permutation` from either the array of images
        `permutation_array` or preimages `inverse_permutation_array`.

        Exactly one argument among `permutation_array` and
        `inverse_permutation_array` (and the deprecated argument `permutation`)
        must be specified.

        The deprecated argument `permutation` should be substituted for
        `inverse_permutation_array`.

        Note that the left action of a permutation on an array `a` is
        `a[inverse_permutation_array]`.

        Args:
            permutation: (deprecated) 1D array listing
                :math:`g^{-1}(x)` for all :math:`0\le x \le N-1`.
            name: Optional, custom name for the permutation.
            permutation_array: 1D array listing
                :math:`g(x)` for all :math:`0\le x \le N-1`.
            inverse_permutation_array: 1D array listing
                :math:`g^{-1}(x)` for all :math:`0\le x \le N-1`.

        Returns:
            A `Permutation` object that encodes the specified permutation.
        """

        arg_list = [permutation, permutation_array, inverse_permutation_array]
        if sum([arg is not None for arg in arg_list]) != 1:
            raise ValueError(
                "Exactly one argument among `permutation`, `permutation_array` "
                "and `inverse_permutation_array` must be specified."
            )

        if permutation is not None:
            warn_deprecation(
                "The argument `permutation` is deprecated.\n\n"
                "In order to clarify notations, you should either pass "
                "the array of images `permutation_array` or "
                "preimages `inverse_permutation_array`."
            )
            inverse_permutation_array = permutation

        if permutation_array is not None:
            inverse_permutation_array = np.argsort(permutation_array)

        self._inverse_permutation_array = HashableArray(
            np.asarray(inverse_permutation_array)
        )

        self.__name = name

    def __hash__(self):
        return hash(self._inverse_permutation_array)

    def __eq__(self, other):
        if isinstance(other, Permutation):
            return np.array_equal(self.permutation_array, other.permutation_array)
        else:
            return False

    @property
    def permutation_array(self):
        return np.asarray(np.argsort(self._inverse_permutation_array))

    @property
    def inverse_permutation_array(self):
        return np.asarray(self._inverse_permutation_array)

    @property
    @deprecated(
        "Deprecated in favor of `permutation.inverse_permutation_array` or "
        "`permutation.permutation_array`"
    )
    def permutation(self):
        return np.asarray(self._inverse_permutation_array)

    @property
    def _name(self):
        return self.__name

    def __repr__(self):
        if self._name is not None:
            return self._name
        else:
            return f"Permutation({self.permutation_array.tolist()})"

    @deprecated_new_name("permutation.inverse_permutation_array")
    def __array__(self, dtype: DType = None):
        return np.asarray(self._inverse_permutation_array, dtype)

    def apply_to_id(self, x: Array):
        """Returns the image of indices `x` under the permutation"""
        return self.permutation_array[x]


@dispatch
def product(p: Permutation, x: Array):
    # if p.permutation is a HashableArray and x is a jax Array
    # direct indexing fails, so we call np.asarray on it to extract the
    # wrapped array
    # TODO make indexing work with HashableArray directly
    return x[..., p.inverse_permutation_array]


@dispatch
def product(p: Permutation, q: Permutation):  # noqa: F811
    name = None if p._name is None and q._name is None else f"{p} @ {q}"
    inverse_permutation_array = q.inverse_permutation_array[p.inverse_permutation_array]
    return Permutation(inverse_permutation_array=inverse_permutation_array, name=name)


@struct.dataclass
class PermutationGroup(FiniteGroup):
    r"""
    Collection of permutation operations acting on sequences of length :code:`degree`.

    Group elements need not all be of type :class:`netket.utils.group.Permutation`,
    only act as such on a sequence when called.

    The class can contain elements that are distinct as objects (e.g.,
    :code:`Identity()` and :code:`Translation((0,))`) but have identical action.
    Those can be removed by calling
    :meth:`~netket.utils.group.PermutationGroup.remove_duplicates`.
    """

    degree: int
    """Number of elements the permutations act on."""

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other: "PermutationGroup") -> bool:
        if isinstance(other, PermutationGroup):
            return np.array_equal(self.to_array(), other.to_array())
        return False

    def _canonical(self, x: Element) -> Array:
        return x(np.arange(self.degree, dtype=int))

    def to_array(self) -> Array:
        r"""
        Convert the abstract group operations to an array of permutation indices.


        It returns a matrix where the `i`-th row contains the indices corresponding
        to the `i`-th group element. That is, :code:`self.to_array()[i, j]`
        is :math:`g_i^{-1}(j)`. Moreover,

        .. code::

            G = # this permutation group...
            V = np.arange(G.degree)
            assert np.all(G(V) == V[..., G.to_array()])

        Returns:
            A matrix that can be used to index arrays in the computational basis
            in order to obtain their permutations.
        """
        return self._canonical_array()

    def __array__(self, dtype=None) -> Array:
        return np.asarray(self.to_array(), dtype=dtype)

    def remove_duplicates(self, *, return_inverse=False) -> "PermutationGroup":
        r"""
        Returns a new :code:`PermutationGroup` with duplicate elements (that is,
        elements which represent identical permutations) removed.

        Args:
            return_inverse: If `True`, also return indices to reconstruct the original
                group from the result.

        Returns:
            The permutation group with duplicate elements removed. If
            :code:`return_inverse==True`, it also returns the indices needed to
            reconstruct the original group from the result.
        """
        if return_inverse:
            group, inverse = super().remove_duplicates(return_inverse=True)
        else:
            group = super().remove_duplicates(return_inverse=False)

        pgroup = PermutationGroup(group.elems, self.degree)

        if return_inverse:
            return pgroup, inverse
        else:
            return pgroup

    @struct.property_cached
    def inverse(self) -> Array:
        try:
            lookup = self._canonical_lookup()
            inverses = []
            # `np.argsort` on a 1D permutation list generates the inverse permutation
            # it acts along last axis by default, so can perform it on to_array()
            # `np.argsort` changes int32 to int64 on Windows,
            # and we need to change it back
            perms = self.to_array()
            invperms = np.argsort(perms).astype(perms.dtype)

            for invperm in invperms:
                inverses.append(lookup[HashableArray(invperm)])

            return np.asarray(inverses, dtype=int)
        except KeyError as err:
            raise RuntimeError(
                "PermutationGroup does not contain the inverse of all elements"
            ) from err

    @struct.property_cached
    def product_table(self) -> Array:
        perms = self.to_array()
        inverse = perms[self.inverse].squeeze()
        n_symm = len(perms)
        lookup = np.unique(np.column_stack((perms, np.arange(len(self)))), axis=0)

        product_table = np.zeros([n_symm, n_symm], dtype=int)
        for i, g_inv in enumerate(inverse):
            row_perms = perms[:, g_inv]
            row_perms = np.unique(
                np.column_stack((row_perms, np.arange(len(self)))), axis=0
            )
            # row_perms should be a permutation of perms, so identical after sorting
            if np.any(row_perms[:, :-1] != lookup[:, :-1]):
                raise RuntimeError(
                    "PermutationGroup is not closed under multiplication"
                )
            # match elements in row_perms to group indices
            product_table[i, row_perms[:, -1]] = lookup[:, -1]

        return product_table

    @property
    def shape(self) -> Shape:
        r"""
        Tuple :code:`(<# of group elements>, <degree>)`.

        Equivalent to :code:`self.to_array().shape`.
        """
        return (len(self), self.degree)

    def apply_to_id(self, x: Array):
        """Returns the image of indices `x` under all permutations"""
        return self.to_array()[self.inverse][:, x]


@dispatch
def product(A: PermutationGroup, B: PermutationGroup):  # noqa: F811
    if A.degree != B.degree:
        raise ValueError(
            "Incompatible groups (`PermutationGroup`s of different degree)"
        )
    return PermutationGroup(
        elems=[a @ b for a, b in itertools.product(A.elems, B.elems)], degree=A.degree
    )


@dispatch
def product(G: PermutationGroup, x: Array):  # noqa: F811
    return np.moveaxis(x[..., G.to_array()], -2, 0)



from functools import partial
from typing import Union

import numpy as np
import sparse

import jax
import jax.numpy as jnp

from flax import struct

from netket.operator import DiscreteJaxOperator
from netket.hilbert import SpinOrbitalFermions
from netket.utils.types import Array
from netket.operator import FermionOperator2nd, FermionOperator2ndJax
from netket.utils.optional_deps import import_optional_dependency

from .._pyscf_utils import (
    TV_from_pyscf_molecule,
    to_desc_order_sparse,
    compute_pyscf_integrals,
)

from .._normal_order_utils import (
    to_normal_order,
    to_normal_order_sector,
    SpinOperatorArrayDict,
)
from ._conversion import (
    fermiop_to_pnc_format_helper,
    fermiop_to_pnc_format_spin_helper,
    pnc_format_to_fermiop_helper,
)
from ._operator_data import (
    collect_ops,
    sparse_arrays_to_coords_data_dict,
    prepare_operator_data_from_coords_data_dict,
    prepare_operator_data_from_coords_data_dict_spin,
    to_coords_data_sector,
    PNCOperatorDataCollectionDict,
    PNCOperatorArrayDict,
    CoordsDataDictSectorCooArrayType,
    CoordsDataDictSectorType,
)
from ._kernels import get_conn_padded_pnc, get_conn_padded_pnc_spin


@struct.dataclass
class ParticleNumberConservingFermioperator2nd(DiscreteJaxOperator):
    r"""
    Particle-number conserving fermionc operator

    .. math::

        \hat H = w + \sum_{ij} w_{ij} \hat c_i^\dagger \hat c_j + \sum_{ijkl} w_{ijkl} \hat c_i^\dagger \hat c_j^\dagger \hat c_k \hat c_l + \sum_{ijklmn} w_{ijklmn} \hat c_i^\dagger \hat c_j^\dagger c_k^\dagger \hat c_l \hat c_m \hat c_n + \dots

    To be used with netket.hilbert.SpinOrbitalFermions with a fixed number of fermions.

    It uses a custom sparse internal representation,
    please refer to the docstrings of prepare_data and prepare_data_diagonal for details.

    We provide several factory methods to create this operator:
        - ParticleNumberConservingFermioperator2nd.from_fermionoperator2nd:
               Conversion form FermionOperator2nd/FermionOperator2ndJax
        - ParticleNumberConservingFermioperator2nd.from_sparse_arrays:
                From sparse arrays (w, w_ij, w_ijkl, w_ijklmn, ...)
        - ParticleNumberConservingFermioperator2nd.from_pyscf_molecule:
                From pyscf

    Furthermore it can be converted to FermionOperator2nd/FermionOperator2ndJax using the .to_fermionoperator2nd() method.
    """

    # factory methods for internal use only:
    # - ParticleNumberConservingFermioperator2nd._from_coords_data_normal_order:
    #         From tuples of (sites, daggers, weights) representing w, w_ij, ...
    #         where only the lower triangular part is nonzero
    # - ParticleNumberConservingFermioperator2nd._from_sparse_arrays_normal_order:
    #         From sparse arrays (w, w_ij, w_ijkl, w_ijklmn) where i>=j, i>=j>=k>=l etc,
    #         and only the lower triangular part is nonzero

    _hilbert: SpinOrbitalFermions = struct.field(pytree_node=False)
    _operator_data: (
        PNCOperatorDataCollectionDict  # custom sparse internal representation
    )

    def get_conn_padded(self, x):
        return get_conn_padded_pnc(self._operator_data, x, self._hilbert.n_fermions)

    @property
    def max_conn_size(self):
        x = jax.ShapeDtypeStruct((1, self._hilbert.size), dtype=jnp.uint8)
        _, mels = jax.eval_shape(self.get_conn_padded, x)
        return mels.shape[-1]

    @property
    def dtype(self):
        x = jax.ShapeDtypeStruct((1, self._hilbert.size), dtype=jnp.uint8)
        _, mels = jax.eval_shape(self.get_conn_padded, x)
        return mels.dtype

    @property
    def is_hermitian(self):
        # TODO more efficient implementation
        return self.to_fermionoperator2nd().is_hermitian

    @classmethod
    def _from_coords_data_normal_order(
        cls,
        hilbert: SpinOrbitalFermions,
        coords_data_dict: PNCOperatorArrayDict,
        **kwargs,
    ):
        r"""
        initialize from PNCOperatorArrayDict

        used internally.
        """
        assert isinstance(hilbert, SpinOrbitalFermions)
        assert hilbert.n_fermions is not None
        n_orbitals = hilbert.n_orbitals * hilbert.n_spin_subsectors
        data = prepare_operator_data_from_coords_data_dict(
            coords_data_dict, n_orbitals, **kwargs
        )
        return cls(hilbert, data)

    @classmethod
    def _from_sparse_arrays_normal_order(
        cls,
        hilbert: SpinOrbitalFermions,
        operators: list[Union[Array, sparse.COO]],
        **kwargs,
    ):
        r"""
        initialize from a list of arrays in normal order (descending)

        Args:
            hilbert: hilbert space
            Operators: list of dense or sparse arrays, each representing an m-body operator for different m

        Example:
        Given an array A of rank 2m with shape (n_orbitals,)*(2m) this initializes the operator

        .. math::

            \hat A = \sum_{i_1 > \dots > i_m, j_1 > \dots > j_m} A_{i_1 \cdots i_m j_1 \cdots j_m} \hat c_{i_1}^\dagger \cdots \hat c_{i_m}^\dagger \hat c_{j_1} \cdots \hat c_{j_m}

        Throws an error if the arrays are not in descending order.
        """
        terms = sparse_arrays_to_coords_data_dict(collect_ops(operators))

        for k, v in terms.items():
            if k <= 2:
                pass
            idx = v[0]
            idx_create = idx[:, : idx.shape[1] // 2]
            idx_destroy = idx[:, idx.shape[1] // 2 :]
            for idx_arr in idx_destroy, idx_create:
                if (jnp.diff(idx_arr) > 0).any():
                    raise ValueError("Input arrays are not in normal order")

        return cls._from_coords_data_normal_order(hilbert, terms, **kwargs)

    @classmethod
    def from_sparse_arrays(
        cls,
        hilbert: SpinOrbitalFermions,
        operators: list[Union[Array, sparse.COO]],
        **kwargs,
    ):
        r"""
        initialize from a list of arrays

        Args:
            hilbert: hilbert space
            operators: list of dense or sparse arrays, each representing an m-body operator for different m
            cutoff: cutoff to use when converting the operators to the internal format.
                    Use a small but nonzero number, to allow for internal equality checks between arrays.

        Example:
        Given an array `A` of rank 2m with shape `(n_orbitals,)*(2m)` this initializes the operator

        .. math::

            \hat A = \sum_{i_1,\dots,i_m, j_1,\dots,j_m} A_{i_1 \cdots i_m j_1 \cdots j_m} \hat c_{i_1}^\dagger \cdots \hat c_{i_m}^\dagger \hat c_{j_1} \cdots \hat c_{j_m}

        """
        # daggers on the left, but not necessarily desc order

        # check shapes
        n_orbitals = hilbert.size
        for op in operators:
            if hasattr(op, "shape") and op.ndim > 0:  # >= 1-body
                s = op.shape
                ndim = op.ndim
                if ndim % 2 != 0:
                    raise ValueError(
                        "operator array has incompatible number of dimensions"
                    )
                if set(s) != {n_orbitals}:
                    raise ValueError(
                        f"inconsistent operator array shapes, {n_orbitals=} but got shape {s}"
                    )

        ops = collect_ops(operators)
        cutoff = kwargs.get("cutoff", 0)
        ops = jax.tree_util.tree_map(partial(to_desc_order_sparse, cutoff=cutoff), ops)
        terms = sparse_arrays_to_coords_data_dict(ops)
        return cls._from_coords_data_normal_order(hilbert, terms, **kwargs)

    @classmethod
    def from_fermionoperator2nd(
        cls, ha: Union[FermionOperator2nd, FermionOperator2ndJax], **kwargs
    ):
        """
        Convert from FermionOperator2nd

        Args:
            ha : the original FermionOperator2nd/FermionOperator2ndJax operator

        Throws an error if the original operator is not particle-number conserving.
        """
        # ha = ha.to_normal_order()
        t = fermiop_to_pnc_format_helper(ha.terms, ha.weights)
        t = to_normal_order(t)
        terms = {k: (v[0], v[2]) for k, v in t.items()}  # drop daggers
        return cls._from_coords_data_normal_order(ha.hilbert, terms, **kwargs)

    def to_fermionoperator2nd(
        self, _cls=FermionOperator2ndJax
    ) -> FermionOperator2ndJax:
        r"""
        Convert to FermionOperator2ndJax
        """
        terms = []
        weights = []
        for d in self._operator_data.values():
            for k, v in d.items():
                t, w = pnc_format_to_fermiop_helper(*v)
                terms = terms + t.tolist()
                weights = weights + w.tolist()
        return _cls(self._hilbert, terms, weights)

    @classmethod
    def from_pyscf_molecule(
        cls,
        mol: "pyscf.gto.mole.Mole",  # noqa: F821
        mo_coeff: Array,
        cutoff: float = 1e-11,
        **kwargs,
    ):
        r"""
        Constructs the operator from a pyscf molecule

        Args:
            mol: pyscf molecule
            mo_coeff: molecular orbital coefficients, e.g. obtained from a HF calculation
            cutoff: cutoff to use when converting the operators to the internal format.
                    Use a small but nonzero number, to allow for internal equality checks between arrays.
        """
        n_orbitals = int(mol.nao)
        hi = SpinOrbitalFermions(n_orbitals, s=1 / 2, n_fermions_per_spin=mol.nelec)
        E_nuc, Tij, Vijkl = TV_from_pyscf_molecule(mol, mo_coeff, cutoff=cutoff)
        return cls._from_sparse_arrays_normal_order(
            hi, [E_nuc, Tij, 0.5 * Vijkl], **kwargs
        )


@struct.dataclass
class ParticleNumberAndSpinConservingFermioperator2nd(DiscreteJaxOperator):
    r"""
    Particle-number conserving and spin-Z-conserving fermionc operator

    .. math::

        \hat H = w + \sum_{ij \sigma} w_{ij \sigma} \hat c_{i \sigma}^\dagger \hat c_{j \sigma} + \sum_{ijkl\sigma \rho} w_{ijkl\sigma\rho}  \hat c_{i\sigma}^\dagger \hat c_{j \rho}^\dagger \hat c_{k \rho} \hat c_{l \sigma}


    Limited to 2-body operators if acting on > 1 sector at a time

    To be used with netket.hilbert.SpinOrbitalFermions with a fixed number of fermions.

    It uses a custom sparse internal representation,
    please refer to the docstrings of prepare_data and prepare_data_diagonal for details.

    We provide several factory methods to create this operator:
        - ParticleNumberAndSpinConservingFermioperator2nd.from_fermionoperator2nd:
                Conversion form FermionOperator2nd/FermionOperator2ndJax (if possible)
        - ParticleNumberConservingFermioperator2nd.from_pyscf_molecule:
                From pyscf
    Furthermore it can be converted to FermionOperator2nd/FermionOperator2ndJax using the .to_fermiop method.
    """

    # factory methods for internal use only:
    # - ParticleNumberAndSpinConservingFermioperator2nd._from_sites_sectors_daggers_weights:
    #         From a dictionary of tuples {k: (sites, sectors, daggers, weights)} representing w, w_ij\sigma, w_ijkl\sigma
    # - ParticleNumberAndSpinConservingFermioperator2nd._from_sparse_arrays_normal_order_all_sectors:
    #         From sparse arrays for w, w_ij and w_ijkl summing over all possible values of \sigma,\rho
    # - ParticleNumberAndSpinConservingFermioperator2nd._from_coords_data:
    #         From a dictionary of tuples {(k, sectors): (sites, daggers, weights)} representing w, w_ij\sigma, w_ijkl\sigma\rho

    _hilbert: SpinOrbitalFermions = struct.field(pytree_node=False)
    _operator_data: PNCOperatorDataCollectionDict

    @property
    def dtype(self):
        x = jax.ShapeDtypeStruct((1, self._hilbert.size), dtype=jnp.uint8)
        _, mels = jax.eval_shape(self.get_conn_padded, x)
        return mels.dtype

    @property
    def is_hermitian(self):
        return NotImplemented

    @property
    def max_conn_size(self):
        x = jax.ShapeDtypeStruct((1, self._hilbert.size), dtype=jnp.uint8)
        _, mels = jax.eval_shape(self.get_conn_padded, x)
        return mels.shape[-1]

    def get_conn_padded(self, x):
        return get_conn_padded_pnc_spin(
            self._operator_data, x, self._hilbert.n_fermions_per_spin
        )

    @classmethod
    def _from_coords_data(
        cls, hilbert: SpinOrbitalFermions, coords_data_sectors: CoordsDataDictSectorType
    ):
        assert isinstance(hilbert, SpinOrbitalFermions)
        assert hilbert.n_fermions is not None
        assert hilbert.n_spin_subsectors >= 2
        n_orbitals = hilbert.n_orbitals
        operator_data = prepare_operator_data_from_coords_data_dict_spin(
            coords_data_sectors, n_orbitals
        )
        return cls(hilbert, operator_data)

    @classmethod
    def _from_sparse_arrays_normal_order(
        cls,
        hilbert: SpinOrbitalFermions,
        operators_sector: CoordsDataDictSectorCooArrayType,
    ):
        r"""
        Args:

            operators_sector:
                dict {key: value}
                    key: a tuple (k, sectors)
                         where k is the number of c/c^\dagger
                         and sectors is a tuple of tuples/numbers containing the index / sets
                         of indices of sectors the operator is acting on each element
                         of sectors needs to be ordered in descending order
                    value: a sparse matrix of coefficeints of shape (n_orbitals,)*k
        """
        coords_data_sectors = sparse_arrays_to_coords_data_dict(operators_sector)
        return cls._from_coords_data(hilbert, coords_data_sectors)

    @classmethod
    def _from_sparse_arrays_normal_order_all_sectors(
        cls,
        hilbert: SpinOrbitalFermions,
        operators: list[Union[Array, sparse.COO]],
        cutoff=1e-11,
    ):
        r"""
        Construct the operator from  sparse arrays for :math:`w, w_ij` and :math:`w_ijkl` summing over all possible values of :math:`\sigma,\rho`,

        .. math::
            \hat H = w + \sum_{ij\sigma} w_{ij} c_{i\sigma}^\dagger c_{j\sigma} + \sigma_{ijkl\sigma\rho} w_{ijkl} c_{i\sigma}^\dagger c_{j\rho}^\dagger c_{k\rho} c_{l\sigma}

        Args:
            hilbert: hilbert space
            operators: a list of sparse arrays representing :math:`w, w_ij` and :math:`w_ijkl`
            cutoff: cutoff on the matirx elements
        """
        # operators = [const, hij, hijkl]
        # ops = {0: const, 2: hij_sparse, 4: hijkl_sparse}
        ops = collect_ops(operators)

        operators_sector = {}
        sectors0 = ()
        sectors1 = tuple(np.arange(hilbert.n_spin_subsectors).tolist())
        sectors2 = tuple(
            map(
                tuple,
                np.array(np.tril_indices(hilbert.n_spin_subsectors, -1)).T.tolist(),
            )
        )
        for k, v in ops.items():
            if k == 0:
                operators_sector[0, sectors0] = v
            elif k == 2:
                operators_sector[2, sectors1] = v
            elif k == 4:
                operators_sector[4, sectors1] = to_desc_order_sparse(v, cutoff)
                # add c_ijkl + c_jilk
                # \sigma_{\sigma!=\rho} c_ijkl  c_i\sigma^\dagger c_j\rho^\dagger c_k\rho c_l\sigma =  \sigma_{\sigma>\rho} (c_ijkl + c_jilk) c_i\sigma^\dagger c_j\rho^\dagger c_k\rho c_l\sigma
                operators_sector[4, sectors2] = v.swapaxes(2, 3) + v.swapaxes(0, 1)
            else:
                raise NotImplementedError
        return cls._from_sparse_arrays_normal_order(hilbert, operators_sector)

    @classmethod
    def from_pyscf_molecule(
        cls,
        mol: "pyscf.gto.mole.Mole",  # noqa: F821
        mo_coeff: Array,
        cutoff: float = 1e-11,
    ):
        r"""
        Constructs the operator from a pyscf molecule

        Args:
            mol: pyscf molecule
            mo_coeff: molecular orbital coefficients, e.g. obtained from a HF calculation
            cutoff: cutoff to use when converting the operators to the internal format.
                    Use a small but nonzero number, to allow for internal equality checks between arrays.
        """

        sparse = import_optional_dependency("sparse")

        n_orbitals = int(mol.nao)
        hilbert = SpinOrbitalFermions(
            n_orbitals, s=1 / 2, n_fermions_per_spin=mol.nelec
        )

        const, hij, hijkl = compute_pyscf_integrals(
            mol, mo_coeff
        )  # not in normal order
        hij = hij * (jnp.abs(hij) > cutoff)
        hij_sparse = sparse.COO.from_numpy(hij)
        hijkl = hijkl * (jnp.abs(hijkl) > cutoff)
        hijkl_sparse = 0.5 * sparse.COO.from_numpy(hijkl)
        return cls._from_sparse_arrays_normal_order_all_sectors(
            hilbert, [const, hij_sparse, hijkl_sparse], cutoff=cutoff
        )

    @classmethod
    def _from_sites_sectors_daggers_weights(
        cls,
        hilbert: SpinOrbitalFermions,
        t: SpinOperatorArrayDict,
        cutoff: float = 1e-11,
    ):
        r"""
        Initialize from SpinOperatorArrayDict
        Args:
            hilbert : hilbert space
            t: { size : (sites, sectors, daggers, weights) }
            cutoff: cutoff to use when converting the operators to the internal format.
                    Use a small but nonzero number, to allow for internal equality checks between arrays.

        Supports arbitrary order of sites, sectors, and daggers,
        it is internally converted to the right order for the operator
        """
        n_orbitals = hilbert.n_orbitals
        n_spin_subsectors = hilbert.n_spin_subsectors
        tno = to_normal_order_sector(t, n_spin_subsectors, n_orbitals)
        coords_data = to_coords_data_sector(
            tno, n_spin_subsectors, n_orbitals, cutoff=cutoff
        )
        return cls._from_coords_data(hilbert, coords_data)

    @classmethod
    def from_fermionoperator2nd(
        cls, ha: Union[FermionOperator2nd, FermionOperator2ndJax], cutoff: float = 1e-11
    ):
        r"""
        Convert from FermionOperator2nd

        Args:
            ha : the original FermionOperator2nd/FermionOperator2ndJax operator
            cutoff: cutoff to use when converting the operators to the internal format.
                    Use a small but nonzero number, to allow for internal equality checks between arrays.

        Throws an error if the original operator is not particle-number conserving, or spin-Z-conserving.
        """
        hilbert = ha.hilbert
        n_orbitals = hilbert.n_orbitals
        n_spin_subsectors = hilbert.n_spin_subsectors
        t = fermiop_to_pnc_format_spin_helper(
            ha.terms, ha.weights, n_orbitals, n_spin_subsectors
        )
        return cls._from_sites_sectors_daggers_weights(hilbert, t, cutoff=cutoff)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import dedent
from typing import Union


import jax
import numpy as np

from netket.utils.dispatch import dispatch

Dim = Union[
    tuple[int], tuple[int, int], tuple[int, int, int], tuple[int, int, int, int]
]


@dispatch
def random_state(hilb, key, *, size=None, dtype=np.float32):
    r"""Generates either a single or a batch of uniformly distributed random states.

    Args:
        hilb: The hilbert space
        key: The Jax :class:`jax.random.KeyArray` generated by :func:`jax.random.PRNGKey`.
        size: If provided, returns a batch of configurations of the form `(size, #)` if
            size is an integer or `(*size, #)` if it is a tuple and where `#` is the Hilbert
            space size. By default, a single random configuration with shape `(#,)` is
            returned.
        dtype: The dtype of the resulting states.

    Example:

        >>> import netket, jax
        >>> hi = netket.hilbert.Qubit(N=2)
        >>> print(hi.random_state(key=jax.random.PRNGKey(0)))
        [1 1]
        >>> print(hi.random_state(size=2, key=jax.random.PRNGKey(1)))
        [[0 1]
         [1 1]]
    """
    return random_state(hilb, key, size, dtype=dtype)


@dispatch
def random_state(hilb, key, size, dtype):  # noqa: F811
    return random_state(hilb, key, size, dtype=dtype)


@dispatch
def random_state(hilb, key, size: None, *, dtype):  # noqa: F811
    return random_state(hilb, key, 1, dtype=dtype)[0]


@dispatch
def random_state(hilb, key, size: Dim, *, dtype):  # noqa: F811
    n = int(np.prod(size))
    return random_state(hilb, key, n, dtype=dtype).reshape(*size, -1)


@dispatch
def random_state(hilb, key, size: int, *, dtype):  # noqa: F811
    raise NotImplementedError(
        dedent(
            f"""
            random_state(hilb, key, size : int, *, dtype) is not implemented for the
            hilbert space {type(hilb)}.

            Define the above function as follows:

            >>>from netket.utils.dispatch import dispatch
            >>>@dispatch
            >>>def random_state(hilb : {type(hilb)}, key, size : int, *, dtype):
            >>>  ...
            >>>  return batch_of_size_states
        """
        )
    )


def flip_state(hilb, key, state, indices):
    r"""
    Given a state `σ` and an index `i`, randomly flips `σ[i]` so that
    `σ_new[i] ≢ σ[i]`.

    Also accepts batched inputs, where state is a batch and indices is a
    vector of ints.

    Returns:
        new_state: a state or batch of states, with one site flipped
        old_vals: a scalar (or vector) of the old values at the flipped sites
    """
    if state.ndim == 1:
        return flip_state_scalar(hilb, key, state, indices)
    else:
        return flip_state_batch(hilb, key, state, indices)


@dispatch
def flip_state_scalar(hilb, key, state, indx):
    new_state, old_val = flip_state_batch(
        hilb, key, state.reshape(1, -1), indx.reshape(1, -1)
    )
    return new_state.reshape(-1), old_val.reshape()


@dispatch
def flip_state_batch(hilb, key, states, indxs):
    keys = jax.random.split(key, states.shape[0])
    res = jax.vmap(flip_state_scalar, in_axes=(None, 0, 0, 0), out_axes=0)(
        hilb, keys, states, indxs
    )
    return res



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING
from functools import partial

import sys

from netket.utils.version_check import module_version

from .base import ModuleFramework, framework

if TYPE_CHECKING:
    from flax import nnx


def _get_graphdef_type(graphdef):
    # TODO: Once we require Jax 0.10.6 (probably once we require jax 0.7)
    # drop this check
    if module_version("flax") >= (0, 10, 6):
        assert len(graphdef.nodes) > 0
        return graphdef.nodes[0].type
    else:
        return graphdef.type


# expose jax-stax as a flax module
class NNXWrapper:
    name: str = "NNX"

    def __init__(self, graphdef):
        self.graphdef = graphdef

    def init(self, rng, *args, **kwargs):
        raise RuntimeError("not allowed")

    def apply(
        self,
        variables,
        *args,
        rngs=None,
        method=None,  # noqa: W0613
        mutable=False,
        **kwargs,
    ):
        if mutable is not False:
            raise NotImplementedError("Not implemented")
        if rngs is not None:
            raise NotImplementedError()
        if method is None:
            method = "__call__"
        elif isinstance(method, partial):
            return method(variables, *args, **kwargs)
        elif not isinstance(method, str):
            raise TypeError(f"method must be a string, not {type(method)}")

        module = self.recompose(variables)
        fun = getattr(module, method)

        return fun(*args, **kwargs)

    def recompose(self, variables):
        from flax import nnx

        model_state = variables["model_state"]
        params = variables["params"]

        nnx_module = nnx.merge(self.graphdef, params, model_state)
        return nnx_module

    def __getattr__(self, name):
        if hasattr(_get_graphdef_type(self.graphdef), name):
            return partial(self.apply, method=name)
        raise AttributeError(
            f"'{type(self).__name__}' (and the wrapped '{self.graphdef.type}') object has no attribute '{name}'"
        )

    def __repr__(self):
        typ = _get_graphdef_type(self.graphdef)
        return f"NNXWrapper(wrapped_class={typ}, ...)"


@framework
class NNXFramework(ModuleFramework):
    name: str = "NNX"

    @property
    def model_contains_parameters(self) -> bool:
        """
        Returns True if the model contains the parameters in the model itself, False
        if the parameters are stored separately.
        """
        return True

    @staticmethod
    def is_loaded() -> bool:
        # this should be not necessary, as netket requires and loads
        # Flax, but let's set a good example
        return "flax" in sys.modules and (
            "flax.experimental.nnx" in sys.modules or "flax.nnx" in sys.modules
        )

    @staticmethod
    def is_my_module(module) -> bool:
        # this will only get called if the module is loaded
        from flax import nnx

        return isinstance(module, nnx.Module) or isinstance(module, NNXWrapper)

    @staticmethod
    def wrap(module: "nnx.Module") -> NNXWrapper:
        from flax import nnx

        if isinstance(module, NNXWrapper):
            return None, module

        graphdef, params, model_state = nnx.split(module, nnx.Param, ...)

        variables = {
            "model_state": model_state.to_pure_dict(),
            "params": params.to_pure_dict(),
        }

        return variables, NNXWrapper(graphdef)

    @staticmethod
    def unwrap(wrapped_module: NNXWrapper, maybe_variables) -> "nnx.Module":
        from flax import nnx

        model_state = maybe_variables["model_state"]
        params = maybe_variables["params"]

        nnx_module = nnx.merge(wrapped_module.graphdef, params, model_state)
        return nnx_module



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Number

import jax
import jax.numpy as jnp
import numpy as np

from . import struct
from .types import DType


@struct.dataclass
class StaticZero(Number):
    """
    A static representation of 0, which can be used in Jax and attempts to behave like
    a normal jax array.

    Used to be able to test for hard-zeros in jax code
    """

    dtype: DType = struct.field(pytree_node=False, default=bool)

    shape = property(lambda _: ())
    ndim = property(lambda _: 0)
    weak_dtype = property(lambda _: True)
    aval = property(lambda self: jax.ShapeDtypeStruct(self.shape, self.dtype))

    __add__ = lambda self, o: self if isinstance(o, StaticZero) else o
    __radd__ = lambda self, o: self if isinstance(o, StaticZero) else o
    __sub__ = lambda self, o: self if isinstance(o, StaticZero) else -o
    __rsub__ = lambda self, o: self if isinstance(o, StaticZero) else o
    __mul__ = lambda self, o: self
    __rmul__ = lambda self, o: self
    __neg__ = lambda self: self
    __bool__ = lambda self: False
    __eq__ = lambda self, o: True if isinstance(o, StaticZero) else False
    astype = lambda self, dtype: StaticZero(dtype)

    def __array__(self, dtype=None):
        if dtype is None:
            dtype = self.dtype
        return np.zeros(self.shape, dtype=dtype)

    def __jax_array__(self, dtype=None):
        if dtype is None:
            dtype = self.dtype
        return jnp.zeros(self.shape, dtype=dtype)

    def __hash__(self):
        return hash(0)



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp
import jax
from functools import partial

from netket.vqs import FullSumState, expect, expect_and_grad
import netket.jax as nkjax
from netket.utils import mpi
from netket.stats import Stats

from .variance_operator import VarianceObservable


@expect.dispatch
def expect(vstate: FullSumState, variance_operator: VarianceObservable):
    if variance_operator.hilbert != vstate.hilbert:
        raise TypeError("Hilbert spaces should match")

    operator_mtrx = variance_operator.operator.to_dense()
    operator_squared_mtrx = variance_operator.operator_squared.to_dense()

    return expect_and_grad_inner_fs(
        vstate._apply_fun,
        vstate.parameters,
        vstate.model_state,
        vstate._all_states,
        operator_mtrx,
        operator_squared_mtrx,
        return_grad=False,
    )


@expect_and_grad.dispatch
def expect_and_grad(
    vstate: FullSumState,
    variance_operator: VarianceObservable,
    *,
    mutable,
):
    if variance_operator.hilbert != vstate.hilbert:
        raise TypeError("Hilbert spaces should match")

    operator_mtrx = variance_operator.operator.to_dense()
    operator_squared_mtrx = variance_operator.operator_squared.to_dense()

    return expect_and_grad_inner_fs(
        vstate._apply_fun,
        vstate.parameters,
        vstate.model_state,
        vstate._all_states,
        operator_mtrx,
        operator_squared_mtrx,
        return_grad=True,
    )


@partial(jax.jit, static_argnames=("afun", "return_grad"))
def expect_and_grad_inner_fs(
    afun, params, model_state, sigma, operator_mtrx, operator_squared_mtrx, return_grad
):
    def expect_kernel_var(params):
        W = {"params": params, **model_state}

        state = jnp.exp(afun(W, sigma))
        state = state / jnp.linalg.norm(state)

        O_mean = state.conj() @ (operator_mtrx @ state)
        O2_mean = state.conj() @ (operator_squared_mtrx @ state)

        return O2_mean - O_mean**2

    if not return_grad:
        var = expect_kernel_var(params)
        return Stats(mean=var, error_of_mean=0.0, variance=0.0)

    var, var_vjp_fun = nkjax.vjp(expect_kernel_var, params, conjugate=True)

    var_grad = var_vjp_fun(jnp.ones_like(var))[0]
    var_grad = jax.tree_util.tree_map(lambda x: mpi.mpi_mean_jax(x)[0], var_grad)

    return Stats(mean=var, error_of_mean=0.0, variance=0.0), var_grad



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import jax
import numpy as np
from jax import numpy as jnp

from netket.utils import config, mpi, struct
from netket.jax.sharding import extract_replicated

from . import mean as _mean
from . import var as _var
from . import total_size as _total_size
from ._autocorr import integrated_time


def _format_decimal(value, std, var):
    if math.isfinite(std) and std > 1e-7:
        decimals = max(int(np.ceil(-np.log10(std))), 0)
        return (
            "{0:.{1}f}".format(value, decimals + 1),
            "{0:.{1}f}".format(std, decimals + 1),
            "{0:.{1}f}".format(var, decimals + 1),
        )
    else:
        return (
            f"{value:.3e}",
            f"{std:.3e}",
            f"{var:.3e}",
        )


_NaN = float("NaN")


def _maybe_item(x):
    if hasattr(x, "shape") and x.shape == ():
        return x.item()
    else:
        return x


@struct.dataclass
class Stats:
    """A dict-compatible pytree containing the result of the statistics function."""

    mean: float | complex = _NaN
    """The mean value."""
    error_of_mean: float = _NaN
    """Estimate of the error of the mean."""
    variance: float = _NaN
    """Estimation of the variance of the data."""
    tau_corr: float = _NaN
    """Estimate of the autocorrelation time (in dimensionless units of number of steps).

    This value is estimated with a blocking algorithm by default, but the result is known
    to be unreliable. A more precise estimator based on the FFT transform can be used by
    setting the environment variable `NETKET_EXPERIMENTAL_FFT_AUTOCORRELATION=1`. This
    estimator is more computationally expensive, but overall the added cost should be
    negligible.
    """
    R_hat: float = _NaN
    """
    Estimator of the split-Rhat convergence estimator.

    The split-Rhat diagnostic is based on comparing intra-chain and inter-chain
    statistics of the sample and is thus only available for 2d-array inputs where
    the rows are independently sampled MCMC chains. In an ideal MCMC samples,
    R_hat should be 1.0. If it deviates from this value too much, this indicates
    MCMC convergence issues. Thresholds such as R_hat > 1.1 or even R_hat > 1.01 have
    been suggested in the literature for when to discard a sample. (See, e.g.,
    Gelman et al., `Bayesian Data Analysis <http://www.stat.columbia.edu/~gelman/book/>`_,
    or Vehtari et al., `arXiv:1903.08008 <https://arxiv.org/abs/1903.08008>`_.)
    """
    tau_corr_max: float = _NaN
    """
    Estimate of the maximum autocorrelation time among all Markov chains.

    This value is only computed if the environment variable
    `NETKET_EXPERIMENTAL_FFT_AUTOCORRELATION` is set.
    """

    def to_dict(self):
        jsd = {}
        jsd["Mean"] = _maybe_item(self.mean)
        jsd["Variance"] = _maybe_item(self.variance)
        jsd["Sigma"] = _maybe_item(self.error_of_mean)
        jsd["R_hat"] = _maybe_item(self.R_hat)
        jsd["TauCorr"] = _maybe_item(self.tau_corr)
        if config.netket_experimental_fft_autocorrelation:
            jsd["TauCorrMax"] = _maybe_item(self.tau_corr_max)
        return jsd

    def to_compound(self):
        return "Mean", self.to_dict()

    def __repr__(self):
        # extract adressable data from fully replicated arrays
        self = extract_replicated(self)
        mean, err, var = _format_decimal(self.mean, self.error_of_mean, self.variance)
        if not math.isnan(self.R_hat):
            ext = f", R̂={self.R_hat:.4f}"
        else:
            ext = ""
        if config.netket_experimental_fft_autocorrelation:
            if not (math.isnan(self.tau_corr) and math.isnan(self.tau_corr_max)):
                ext += f", τ={self.tau_corr:.1f}<{self.tau_corr_max:.1f}"
        return f"{mean} ± {err} [σ²={var}{ext}]"

    # Alias accessors
    def __getattr__(self, name):
        if name in ("mean", "Mean"):
            return self.mean
        elif name in ("variance", "Variance"):
            return self.variance
        elif name in ("error_of_mean", "Sigma"):
            return self.error_of_mean
        elif name in ("R_hat", "R"):
            return self.R_hat
        elif name in ("tau_corr", "TauCorr"):
            return self.tau_corr
        elif name in ("tau_corr_max", "TauCorrMax"):
            return self.tau_corr_max
        else:
            raise AttributeError(f"'Stats' object object has no attribute '{name}'")

    def real(self):
        return self.replace(mean=np.real(self.mean))

    def imag(self):
        return self.replace(mean=np.imag(self.mean))


def _get_blocks(data, block_size):
    chain_length = data.shape[1]

    n_blocks = int(np.floor(chain_length / float(block_size)))

    return data[:, 0 : n_blocks * block_size].reshape((-1, block_size)).mean(axis=1)


def _block_variance(data, l):
    blocks = _get_blocks(data, l)
    ts = _total_size(blocks)
    if ts > 0:
        return _var(blocks), ts
    else:
        return jnp.nan, 0


def _batch_variance(data):
    b_means = data.mean(axis=1)
    ts = _total_size(b_means)
    return _var(b_means), ts


def _split_R_hat(data, W):
    N = data.shape[-1]
    if not config.netket_use_plain_rhat:
        # compute split-chain batch variance
        local_batch_size = data.shape[0]
        if N % 2 == 0:
            # split each chain in the middle,
            # like [[1 2 3 4]] -> [[1 2][3 4]]
            batch_var, _ = _batch_variance(data.reshape(2 * local_batch_size, N // 2))
        else:
            # drop the last sample of each chain for an even split,
            # like [[1 2 3 4 5]] -> [[1 2][3 4]]
            batch_var, _ = _batch_variance(
                data[:, :-1].reshape(2 * local_batch_size, N // 2)
            )

    # V_loc = _np.var(data, axis=-1, ddof=0)
    # W_loc = _np.mean(V_loc)
    # W = _mean(W_loc)
    # # This approximation seems to hold well enough for larger n_samples
    return jnp.sqrt((N - 1) / N + batch_var / W)


BLOCK_SIZE = 32


def statistics(data: jax.Array) -> Stats:
    r"""
    Returns statistics of a given array (or matrix, see below) containing a stream of data.
    This is particularly useful to analyze Markov Chain data, but it can be used
    also for other type of time series.
    Assumes same shape on all MPI processes.

    Args:
        data (vector or matrix): The input data. It can be real or complex valued.
            * if a vector, it is assumed that this is a time series of data (not necessarily independent);
            * if a matrix, it is assumed that that rows :code:`data[i]` contain independent time series.

    Returns:
       Stats:
        A dictionary-compatible class containing the
        average (:code:`.mean`, :code:`["Mean"]`),
        variance (:code:`.variance`, :code:`["Variance"]`),
        the Monte Carlo standard error of the mean (:code:`error_of_mean`, :code:`["Sigma"]`),
        an estimate of the autocorrelation time (:code:`tau_corr`, :code:`["TauCorr"]`), and the
        Gelman-Rubin split-Rhat diagnostic (:code:`.R_hat`, :code:`["R_hat"]`).

        If the flag `NETKET_EXPERIMENTAL_FFT_AUTOCORRELATION` is set, the autocorrelation is computed
        exactly using a FFT transform, and an extra field `tau_corr_max` is inserted in the
        statistics object

        These properties can be accessed both the attribute and the dictionary-style syntax
        (both indicated above).

        The split-Rhat diagnostic is based on comparing intra-chain and inter-chain
        statistics of the sample and is thus only available for 2d-array inputs where
        the rows are independently sampled MCMC chains. In an ideal MCMC samples,
        R_hat should be 1.0. If it deviates from this value too much, this indicates
        MCMC convergence issues. Thresholds such as R_hat > 1.1 or even R_hat > 1.01 have
        been suggested in the literature for when to discard a sample. (See, e.g.,
        Gelman et al., `Bayesian Data Analysis <http://www.stat.columbia.edu/~gelman/book/>`_,
        or Vehtari et al., `arXiv:1903.08008 <https://arxiv.org/abs/1903.08008>`_.)
    """
    if config.netket_experimental_fft_autocorrelation:
        return _statistics(data)
    else:
        from .mc_stats_old import statistics as statistics_blocks

        return statistics_blocks(data)


@jax.jit
def _statistics(data):
    data = jnp.atleast_1d(data)
    if data.ndim == 1:
        data = data.reshape((1, -1))

    if data.ndim > 2:
        raise NotImplementedError("Statistics are implemented only for ndim<=2")

    mean = _mean(data)
    variance = _var(data)

    taus = jax.vmap(integrated_time)(data)
    tau_avg, _ = mpi.mpi_mean_jax(jnp.mean(taus))
    tau_max, _ = mpi.mpi_max_jax(jnp.max(taus))

    batch_var, n_batches = _batch_variance(data)
    if n_batches > 1:
        error_of_mean = jnp.sqrt(batch_var / n_batches)
        R_hat = _split_R_hat(data, variance)
    else:
        l_block = max(1, data.shape[1] // BLOCK_SIZE)
        block_var, n_blocks = _block_variance(data, l_block)
        error_of_mean = jnp.sqrt(block_var / n_blocks)
        R_hat = jnp.nan

    res = Stats(mean, error_of_mean, variance, tau_avg, R_hat, tau_max)

    return res



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from netket.vqs import FullSumState, expect
from netket.stats import Stats
from netket.utils import module_version

from .S2_operator import Renyi2EntanglementEntropy


@expect.dispatch
def Renyi2(vstate: FullSumState, op: Renyi2EntanglementEntropy):
    if op.hilbert != vstate.hilbert:
        raise TypeError("Hilbert spaces should match")

    state_qutip = vstate.to_qobj()

    N = vstate.hilbert.size

    if len(op.partition) == N or len(op.partition) == 0:
        out = 0
    else:
        mask = np.zeros(N, dtype=bool)
        mask[op.partition] = True

        rdm = state_qutip.ptrace(np.arange(N)[mask])

        # TODO: when we drop support for qutip older versions, always keep this
        # qutip 5 was released in march 2024
        if module_version("qutip") >= (5, 0, 0):
            rdm = rdm.data.to_array()

        n = 2
        out = np.log2(np.trace(np.linalg.matrix_power(rdm, n))) / (1 - n)
        out = np.absolute(out.real)

    S2_stats = Stats(mean=out, error_of_mean=0.0, variance=0.0)

    return S2_stats



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Ignore false-positives for redefined `product` functions:
# pylint: disable=function-redefined

import itertools
from functools import partial
from math import pi

import numpy as np
from scipy.linalg import schur

from netket.utils import HashableArray, struct
from netket.utils.float import comparable, comparable_periodic, is_approx_int
from netket.utils.types import Array, Shape
from netket.utils.dispatch import dispatch

from ._group import FiniteGroup
from ._semigroup import Element, Identity

############ POINT GROUP SYMMETRY CLASS ########################################


@struct.dataclass
class PGSymmetry(Element):
    """
    An abstract group element object to geometrically describe point group symmetries.

    Construction: `PGSymmetry(W,w)` with an orthogonal matrix `W` and an optional
    translation vector `w` returns an object that maps vectors :math:`\vec x` to
    :math:`W\vec x + \vec w`.

    Internally, the transformation is stored as the affine matrix [[W,w],[0,1]],
    such that matrix inverses and multiplication correspond to inverting and
    multiplying the transformation itself.
    """

    _affine: Array
    """
    A 2D array specifying the affine transformation. It has to be of the block
    structure [[W,w],[0,1]].
    """

    def __pre_init__(self, W: Array, w: Array | None = None) -> tuple[tuple, dict]:
        W = np.asarray(W)
        if W.ndim != 2 or W.shape[0] != W.shape[1]:
            raise ValueError("W must be a 2D square matrix")
        ndim = W.shape[0]
        if not np.allclose(W.T @ W, np.eye(ndim)):
            raise ValueError("W must be an orthogonal matrix")
        if w is None:
            w = np.zeros((ndim, 1))
        else:
            w = np.asarray(w).reshape(ndim, 1)
        return (np.block([[W, w], [np.zeros(ndim), 1]]),), {}

    @property
    def affine_matrix(self) -> Array:
        """
        Returns the (d+1) × (d+1) dimensional matrix representing the action of
        `self` as an affine transformation.
        """
        return self._affine

    @property
    def matrix(self) -> Array:
        """
        Returns the d×d dimensional matrix representing the rotational action
        of `self`.
        """
        return self._affine[:-1, :-1]

    @property
    def translation(self) -> Array:
        """Returns the translation vector associated with `self`."""
        return self._affine[:-1, -1]

    @property
    def ndim(self) -> int:
        """Returns the dimension of vectors `self` acts on."""
        return self._affine.shape[0] - 1

    @property
    def is_proper(self) -> bool:
        """Returns True if `self` is a proper rotation (det(`self.matrix`) is +1)."""
        return np.isclose(np.linalg.det(self.matrix), 1.0)

    @property
    def is_symmorphic(self) -> bool:
        """Returns True if `self` leaves the origin in place."""
        return np.allclose(self.translation, 0.0)

    def preimage(self, x):
        """
        Returns the preimage of `x` under the transformation
        (i.e., `self @ self.preimage(x) == x` up to numerical accuracy.
        """
        return np.tensordot(x - self.translation, self.matrix, axes=1)

    def k_action(self, x):
        """
        Returns the action of `self` on the input without the associated translation.
        This is how the symmetry acts on wave vectors.
        """
        return np.tensordot(x, self.matrix.T, axes=1)

    def __hash__(self):
        return hash(HashableArray(comparable(self._affine)))

    def __eq__(self, other):
        if isinstance(other, PGSymmetry):
            return HashableArray(comparable(self._affine)) == HashableArray(
                comparable(other._affine)
            )
        else:
            return False

    def change_origin(self, origin: Array) -> "PGSymmetry":
        """Returns a `PGSymmetry` representing a pure point-group transformation
        around `origin` with transformation matrix `self._W`."""
        return PGSymmetry(
            self.matrix, (np.eye(self.ndim) - self.matrix) @ np.asarray(origin)
        )

    @struct.property_cached
    def _name(self) -> str:
        if self.ndim == 2:
            return _2D_name(self.matrix, self.translation)
        elif self.ndim == 3:
            return _3D_name(self.matrix, self.translation)
        else:
            return f"PGSymmetry({self.matrix}, {self.translation})"

    def __repr__(self) -> str:
        return self._name  # noqa: E0306


@dispatch
def product(p: PGSymmetry, x: Array):  # noqa: F811
    return np.tensordot(x, p.matrix.T, axes=1) + p.translation


@dispatch
def product(p: PGSymmetry, q: PGSymmetry):  # noqa: F811
    return PGSymmetry(p.matrix @ q.matrix, p.matrix @ q.translation + p.translation)


############ NAMING 2D AND 3D POINT GROUP SYMMETRIES ###########################

_naming_tol = 1e-6
_naming_allclose = partial(np.allclose, atol=_naming_tol, rtol=0.0)
_naming_isclose = partial(np.isclose, atol=_naming_tol, rtol=0.0)


# use Schur decomposition for eigenvalues of orthogonal W matrices to ensure
# that eigenvectors are always orthogonal
def _eig(W):
    e, v = schur(W, "complex")
    return np.diag(e), v


def _origin_trans(W: Array, w: Array) -> tuple[Array, Array]:
    """Decomposes a point group symmetry into a pure (improper) rotation around
    an origin and a translation along the axis/plane of the transformation.
    Returns the tuple (origin, translation)."""
    e, v = _eig(np.eye(W.shape[0]) - W)
    # eigenvectors with eigenvalue 1 correspond to translations
    trans_v = v[:, _naming_isclose(e, 0.0)]
    trans = trans_v @ trans_v.T.conj() @ w
    # eigenvectors with other eigenvalues allow shifting the origin
    origin_v = v[:, np.logical_not(_naming_isclose(e, 0.0))]
    origin_e = np.diag(1 / e[np.logical_not(_naming_isclose(e, 0.0))])
    origin = origin_v @ origin_e @ origin_v.T.conj() @ w
    return origin.real, trans.real


def _2D_name(W: Array, w: Array | None) -> str:
    if W.shape != (2, 2):
        raise ValueError("This function names 2D symmetries")

    if w is None:
        origin = trans = np.zeros(2)
    else:
        origin, trans = _origin_trans(W, w)

    origin = "" if _naming_allclose(origin, 0.0) else f"O{_to_rational_vector(origin)}"
    trans = None if _naming_allclose(trans, 0.0) else f"{_to_rational_vector(trans)}"

    if _naming_isclose(np.linalg.det(W), 1.0):  # rotations
        if _naming_allclose(W, np.eye(2)):  # identity / translation
            if trans is None:
                return "Id()"
            else:
                return f"Translation{trans}"
        else:
            angle = np.arctan2(W[1, 0], W[0, 0])
            # in practice, all rotations are by integer degrees
            angle = int(np.rint(np.degrees(angle)))
            return f"Rot({angle}°){origin}"

    elif _naming_isclose(np.linalg.det(W), -1.0):  # reflections / glides
        axis = np.arctan2(W[1, 0], W[0, 0]) / 2
        axis = int(np.rint(np.degrees(axis)))
        if trans is None:
            return f"Refl({axis}°){origin}"
        else:
            return f"Glide{trans}{origin}"

    else:
        raise ValueError("W must be an orthogonal matrix")


def _3D_name(W: Array, w: Array | None = None) -> str:
    if W.shape != (3, 3):
        raise ValueError("This function names 3D symmetries")

    if w is None:
        origin = trans = np.zeros(3)
    else:
        origin, trans = _origin_trans(W, w)

    origin = "" if _naming_allclose(origin, 0.0) else f"O{_to_rational_vector(origin)}"

    if _naming_isclose(np.linalg.det(W), 1.0):  # rotations / screws
        if _naming_allclose(W, np.eye(3)):  # identity / translation
            if _naming_allclose(trans, 0.0):
                return "Id()"
            else:
                return f"Translation{_to_rational_vector(trans)}"

        else:  # actual rotations / screws
            e, v = _eig(W)

            if _naming_isclose(np.trace(W), -1.0):  # π-rotations
                angle = pi
                # rotation axis is eigenvector with eigenvalue +1
                axis = v[:, _naming_isclose(e, 1.0)].real.flatten()

            else:  # pick axis s.t. rotation angle be positive
                pos = e.imag > _naming_tol
                angle = np.angle(e[pos])[0]
                v = v[:, pos].flatten()
                axis = np.cross(v.imag, v.real)
                if not _naming_allclose(trans, 0.0):
                    # screws may have negative angles if trans, axis are opposite
                    angle *= np.sign(axis @ trans)

            angle = int(np.rint(np.degrees(angle)))
            if _naming_allclose(trans, 0.0):
                return f"Rot({angle}°){_to_int_vector(axis)}{origin}"
            else:
                return f"Screw({angle}°){_to_rational_vector(trans)}{origin}"

    elif _naming_isclose(np.linalg.det(W), -1.0):  # improper rotations
        if _naming_allclose(W, -np.eye(3)):  # inversion
            return f"Inv(){origin}"

        elif _naming_isclose(np.trace(W), 1.0):  # reflections / glides
            e, v = _eig(W)
            # reflection plane normal is eigenvector with eigenvalue -1
            axis = v[:, _naming_isclose(e, -1.0)].real.flatten()
            # convention: first nonzero entry is positive
            axis *= np.sign(axis[np.logical_not(_naming_isclose(axis, 0.0))][0])
            if _naming_allclose(trans, 0.0):
                return f"Refl{_to_int_vector(axis)}{origin}"
            else:
                return (
                    f"Glide{_to_rational_vector(trans)}ax{_to_int_vector(axis)}{origin}"
                )

        else:  # rotoreflections, choose axis s.t. rotation angle be positive
            e, v = _eig(W)
            pos = e.imag > _naming_tol
            angle = np.angle(e[pos])[0]
            angle = int(np.rint(np.degrees(angle)))
            v = v[:, pos].flatten()
            axis = np.cross(v.imag, v.real)
            return f"RotoRefl({angle}°){_to_int_vector(axis)}{origin}"

    else:
        raise ValueError("W must be an orthogonal matrix")


def __to_rational(x: float) -> tuple[int, int]:
    denom = is_approx_int(x * np.arange(1, 100), atol=_naming_tol)
    if not denom.any():
        raise ValueError
    denom = np.arange(1, 100)[denom][0]
    return int(np.rint(x * denom)), denom


def _to_rational(x: float) -> str:
    try:
        numer, denom = __to_rational(x)
        return f"{numer}/{denom}" if denom != 1 else f"{numer}"
    except ValueError:
        # in hexagonal symmetry, you often get a √3 in the x/y coordinate
        try:
            numer, denom = __to_rational(x / 3**0.5)
            numer = "" if numer == 1 else ("-" if numer == -1 else numer)
            denom = "" if denom == 1 else f"/{denom}"
            return f"{numer}√3{denom}"
        except ValueError:
            # just return it as it is
            return f"{x:.3f}"


def _to_rational_vector(v: Array) -> Array:
    return "[" + ",".join(map(_to_rational, v)) + "]"


def __to_int_vector(v: Array) -> Array:
    # if there is a small integer representation, v/sum(abs(v)) consists of
    # rationals with small denominators
    v = v / np.abs(v).sum()
    scaled_v = np.outer(np.arange(1, 100), v)  # upper cutoff is arbitrary
    is_int_vector = np.all(_naming_isclose(scaled_v, np.rint(scaled_v)), axis=1)
    if np.any(is_int_vector):
        scaled_v = scaled_v[is_int_vector][0]
        return np.asarray(np.rint(scaled_v), dtype=int)
    else:
        raise ValueError


def _to_int_vector(v: Array) -> str:
    try:
        v = __to_int_vector(v)
        return f"[{v[0]},{v[1]},{v[2]}]"
    except ValueError:
        # in hexagonal symmetry, you often get a √3 in the x/y coordinate
        try:
            w = v.copy()
            w[1] /= 3**0.5
            w = __to_int_vector(w)
            return f"[{w[0]},{w[1]}√3,{w[2]}]"
        except ValueError:
            # just return a normalised v
            v = v / np.linalg.norm(v)
            return f"[{v[0]:.3f},{v[1]:.3f},{v[2]:.3f}]"


############ POINT GROUP CLASS #################################################


@struct.dataclass()
class PointGroup(FiniteGroup):
    """
    Collection of point group symmetries acting on n-dimensional vectors.

    Group elements need not all be of type :class:`netket.utils.symmetry.PGSymmetry`,
    only act on such vectors when called. Currently, however, only `Identity` and
    `PGSymmetry` have canonical forms implemented.

    The class can contain elements that are distinct as objects (e.g.,
    :code:`Identity()` and :code:`Rotation(0)`) but have identical action.
    Those can be removed by calling :code:`remove_duplicates`.
    """

    ndim: int
    """Dimensionality of point group operations."""

    unit_cell: Array | None = None
    """Lattice vectors of the unit cell on which the point group acts.
    It is used to match non-symmorphic symmetries modulo lattice translations."""

    def __hash__(self):
        return super().__hash__()

    @struct.property_cached
    def is_symmorphic(self) -> bool:
        return all(
            [
                (True if isinstance(elem, Identity) else elem.is_symmorphic)
                for elem in self.elems
            ]
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, PointGroup):
            return False
        if (self.unit_cell is None) != (other.unit_cell is None):
            return False
        if not np.allclose(self.unit_cell, other.unit_cell):
            return False
        # for simplicity we also require that they be ordered the same way
        for i, elem in enumerate(self.elems):
            if elem != other.elems[i]:
                return False
        return True

    def matrix(self, x: Element) -> Array:
        if isinstance(x, Identity):
            return np.eye(self.ndim, dtype=float)
        elif isinstance(x, PGSymmetry):
            return x.matrix
        else:
            raise ValueError(
                "`PointGroup` only supports `Identity` and `PGSymmetry` elements"
            )

    def translation(self, x: Element) -> Array:
        if isinstance(x, Identity):
            return np.zeros(self.ndim, dtype=float)
        elif isinstance(x, PGSymmetry):
            return x.translation
        else:
            raise ValueError(
                "`PointGroup` only supports `Identity` and `PGSymmetry` elements"
            )

    def affine_matrix(self, x: Element) -> Array:
        if isinstance(x, Identity):
            return np.eye(self.ndim + 1, dtype=float)
        elif isinstance(x, PGSymmetry):
            return x.affine_matrix
        else:
            raise ValueError(
                "`PointGroup` only supports `Identity` and `PGSymmetry` elements"
            )

    def _canonical_from_affine_matrix(self, affine: Array) -> Array:
        if self.unit_cell is None:
            return comparable(affine)
        else:
            return np.vstack(
                (
                    comparable(affine[0 : self.ndim, 0 : self.ndim]),
                    comparable_periodic(
                        affine[0 : self.ndim, self.ndim] @ np.linalg.inv(self.unit_cell)
                    ),
                )
            )

    def _canonical(self, x: Element) -> Array:
        return self._canonical_from_affine_matrix(self.affine_matrix(x))

    def to_array(self) -> Array:
        """
        Convert the abstract group operations to an array of transformation matrices

        For symmorphic groups, `self.to_array()[i]` contains the transformation
        matrix of the `i`th group element.
        For nonsymmorphic groups, `self.to_array()[i]` is a (d+1)×(d+1) block
        matrix of the form [[W,w],[0,1]]: multiplying these matrices is
        equivalent to multiplying the symmetry operations.
        """
        return np.asarray([self.affine_matrix(x) for x in self.elems])

    def matrices(self) -> Array:
        return np.asarray([self.matrix(x) for x in self.elems])

    def translations(self) -> Array:
        return np.asarray([self.translation(x) for x in self.elems])

    def __array__(self, dtype=None) -> Array:
        return np.asarray(self.to_array(), dtype=dtype)

    def remove_duplicates(self, *, return_inverse=False) -> "PointGroup":
        """
        Returns a new :code:`PointGroup` with duplicate elements (that is, elements
        which represent identical transformations) removed.

        Arguments:
            return_inverse: If True, also return indices to reconstruct the original
                group from the result.

        Returns:
            The point group with duplicate elements removed. If
            :code:`return_inverse==True`, then it also returns the list
            of indices needed to reconstruct the original group from the result.
        """
        if return_inverse:
            group, inverse = super().remove_duplicates(return_inverse=True)
        else:
            group = super().remove_duplicates(return_inverse=False)

        pgroup = PointGroup(group.elems, self.ndim, unit_cell=self.unit_cell)

        if return_inverse:
            return pgroup, inverse
        else:
            return pgroup

    def rotation_group(self) -> "PointGroup":
        """
        Returns a new `PointGroup` that represents the subgroup of rotations
        (i.e. symmetries where the determinant of the transformation matrix is +1)
        in `self`
        """
        subgroup = []
        for i in self.elems:
            if isinstance(i, Identity):
                subgroup.append(i)
            elif i.is_proper:
                subgroup.append(i)
        return PointGroup(subgroup, self.ndim, unit_cell=self.unit_cell)

    def change_origin(self, origin: Array) -> "PointGroup":
        """
        Returns a new `PointGroup`, `out`, such that all elements of `out`
        describe pure point-group transformations around `origin` and `out[i]`
        has the same transformation matrix as `self[i]`.
        """
        out: list[Element] = []
        for elem in self.elems:
            if isinstance(elem, Identity):
                out.append(Identity())
            else:
                out.append(elem.change_origin(origin))
        return PointGroup(out, self.ndim, unit_cell=self.unit_cell)

    @struct.property_cached
    def inverse(self) -> Array:
        try:
            lookup = self._canonical_lookup()
            affine_matrices = self.to_array()

            inverse = np.zeros(len(self.elems), dtype=int)

            for index in range(len(self)):
                inverse_matrix = np.linalg.inv(affine_matrices[index])
                inverse[index] = lookup[
                    HashableArray(self._canonical_from_affine_matrix(inverse_matrix))
                ]

            return inverse
        except KeyError as err:
            raise RuntimeError(
                "PointGroup does not contain the inverse of all elements"
            ) from err

    @struct.property_cached
    def product_table(self) -> Array:
        try:
            # again, we calculate the product table of transformation matrices directly
            affine_matrices = self.to_array()
            product_matrices = np.einsum(
                "iab, jbc -> ijac", affine_matrices, affine_matrices
            )  # this is a table of M_g M_h

            lookup = self._canonical_lookup()

            n_symm = len(self)
            product_table = np.zeros((n_symm, n_symm), dtype=int)

            for i in range(n_symm):
                for j in range(n_symm):
                    product_table[i, j] = lookup[
                        HashableArray(
                            self._canonical_from_affine_matrix(product_matrices[i, j])
                        )
                    ]

            return product_table[self.inverse]  # reshuffle rows to match specs
        except KeyError as err:
            raise RuntimeError("PointGroup is not closed under multiplication") from err

    @property
    def shape(self) -> Shape:
        """
        Tuple `(<# of group elements>, <ndim>, <ndim>)`.
        Equivalent to :code:`self.to_array().shape`.
        """
        return (len(self), self.ndim + 1, self.ndim + 1)


def trivial_point_group(ndim: int) -> PointGroup:
    return PointGroup([Identity()], ndim=ndim)


@dispatch
def product(A: PointGroup, B: PointGroup):  # noqa: F811
    if A.ndim != B.ndim:
        raise ValueError("Incompatible groups (`PointGroup`s of different dimension)")
    if A.unit_cell is None:
        unit_cell = B.unit_cell
    else:
        if B.unit_cell is not None and not np.allclose(A.unit_cell, B.unit_cell):
            raise ValueError(
                "Incompatible groups (`PointGroup`s with different unit cells)"
            )
        unit_cell = A.unit_cell
    return PointGroup(
        elems=[a @ b for a, b in itertools.product(A.elems, B.elems)],
        ndim=A.ndim,
        unit_cell=unit_cell,
    )


@dispatch
def product(G: PointGroup, x: Array):  # noqa: F811
    return np.moveaxis(
        np.tensordot(x, G.matrices().transpose(2, 0, 1), axes=1) + G.translations(),
        -2,
        0,
    )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import sqrt
from typing import Any
from collections.abc import Callable

from jax import numpy as jnp
from jax.nn.initializers import zeros

from netket.models.autoreg import ARNNSequential, _get_feature_list, _normalize
from netket.nn import FastMaskedConv1D, FastMaskedConv2D, FastMaskedDense1D
from netket.nn import activation as nkactivation
from netket.nn.masked_linear import default_kernel_init
from netket.utils.types import Array, DType, NNInitFunc


class FastARNNSequential(ARNNSequential):
    """
    Implementation of a fast ARNN that sequentially calls its layers and activation function.

    Subclasses must implement `activation` as a field or a method,
    and assign a list of fast ARNN layers to `self._layers` in `setup`.

    The fast autoregressive sampling is described in `Ramachandran et. {\\it al} <https://arxiv.org/abs/1704.06001>`_.
    To generate one sample using an autoregressive network, we need to evaluate the network `N` times, where `N` is
    the number of input sites. But actually we only change one input site each time, and not all intermediate results
    depend on the changed input because of the autoregressive property, so we can cache unchanged intermediate results
    and avoid repeated computation.

    This optimization is particularly useful for convolutional neural networks (CNN) and recurrent neural networks (RNN)
    where each output site of a layer only depends on a small number of input sites, while not so useful for densely
    connected layers.
    """

    def conditional(self, inputs: Array, index: int) -> Array:
        """
        Computes the conditional probabilities for one site to take each value.
        See `AbstractARNN.conditional`.
        """
        if inputs.ndim == 1:
            inputs = jnp.expand_dims(inputs, axis=0)

        x = jnp.expand_dims(inputs, axis=-1)
        x = self._take_prev_site(x, index)

        for i in range(len(self._layers)):
            if i > 0 and hasattr(self, "activation"):
                x = self.activation(x)
            x = self._layers[i].update_site(x, index)

        log_psi = _normalize(x, self.machine_pow)
        p = jnp.exp(self.machine_pow * log_psi.real)
        return p

    def _take_prev_site(self, inputs: Array, index: int) -> Array:
        """
        Takes the previous site in the autoregressive order.
        """
        # When `index = 0`, it doesn't matter which site we take
        return inputs[:, index - 1]


class FastARNNDense(FastARNNSequential):
    """
    Fast autoregressive neural network with dense layers.

    See :class:`netket.models.FastARNNSequential` for a brief explanation
    of fast autoregressive sampling.

    TODO: FastMaskedDense1D does not support JIT yet, because it involves
    slicing the cached inputs and the weights with a dynamic shape.
    """

    layers: int
    """number of layers."""
    features: tuple[int, ...] | int
    """output feature density in each layer. If a single number is given,
    all layers except the last one will have the same number of features."""
    activation: Callable[[Array], Array] = nkactivation.reim_selu
    """the nonlinear activation function between hidden layers (default: reim_selu)."""
    use_bias: bool = True
    """whether to add a bias to the output (default: True)."""
    param_dtype: DType = jnp.float64
    """the dtype of the computation (default: float64)."""
    precision: Any = None
    """numerical precision of the computation, see :class:`jax.lax.Precision` for details."""
    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the weights."""
    bias_init: NNInitFunc = zeros
    """initializer for the biases."""
    machine_pow: int = 2
    """exponent to normalize the outputs of `__call__`."""

    def setup(self):
        features = _get_feature_list(self)
        self._layers = [
            FastMaskedDense1D(
                size=self.hilbert.size,
                features=features[i],
                exclusive=(i == 0),
                use_bias=self.use_bias,
                param_dtype=self.param_dtype,
                precision=self.precision,
                kernel_init=self.kernel_init,
                bias_init=self.bias_init,
            )
            for i in range(self.layers)
        ]


class FastARNNConv1D(FastARNNSequential):
    """
    Fast autoregressive neural network with 1D convolution layers.

    See :class:`netket.models.FastARNNSequential` for a brief
    explanation of fast autoregressive sampling.
    """

    layers: int
    """number of layers."""
    features: tuple[int, ...] | int
    """output feature density in each layer. If a single number is given,
    all layers except the last one will have the same number of features."""
    kernel_size: int
    """length of the convolutional kernel."""
    kernel_dilation: int = 1
    """dilation factor of the convolution kernel (default: 1)."""
    activation: Callable[[Array], Array] = nkactivation.reim_selu
    """the nonlinear activation function between hidden layers (default: reim_selu)."""
    use_bias: bool = True
    """whether to add a bias to the output (default: True)."""
    param_dtype: DType = jnp.float64
    """the dtype of the computation (default: float64)."""
    precision: Any = None
    """numerical precision of the computation, see :class:`jax.lax.Precision` for details."""
    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the weights."""
    bias_init: NNInitFunc = zeros
    """initializer for the biases."""
    machine_pow: int = 2
    """exponent to normalize the outputs of `__call__`."""

    def setup(self):
        features = _get_feature_list(self)
        self._layers = [
            FastMaskedConv1D(
                features=features[i],
                kernel_size=self.kernel_size,
                kernel_dilation=self.kernel_dilation,
                exclusive=(i == 0),
                use_bias=self.use_bias,
                param_dtype=self.param_dtype,
                precision=self.precision,
                kernel_init=self.kernel_init,
                bias_init=self.bias_init,
            )
            for i in range(self.layers)
        ]


class FastARNNConv2D(FastARNNSequential):
    """
    Fast autoregressive neural network with 2D convolution layers.

    See :class:`netket.models.FastARNNSequential` for a brief explanation of fast autoregressive sampling.
    """

    layers: int
    """number of layers."""
    features: tuple[int, ...] | int
    """output feature density in each layer. If a single number is given,
    all layers except the last one will have the same number of features."""
    kernel_size: tuple[int, int]
    """shape of the convolutional kernel `(h, w)`. Typically, `h = w // 2 + 1`."""
    kernel_dilation: tuple[int, int] = (1, 1)
    """a sequence of 2 integers, giving the dilation factor to
    apply in each spatial dimension of the convolution kernel (default: 1)."""
    activation: Callable[[Array], Array] = nkactivation.reim_selu
    """the nonlinear activation function between hidden layers (default: reim_selu)."""
    use_bias: bool = True
    """whether to add a bias to the output (default: True)."""
    param_dtype: DType = jnp.float64
    """the dtype of the computation (default: float64)."""
    precision: Any = None
    """numerical precision of the computation, see :class:`jax.lax.Precision` for details."""
    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the weights."""
    bias_init: NNInitFunc = zeros
    """initializer for the biases."""
    machine_pow: int = 2
    """exponent to normalize the outputs of `__call__`."""

    def setup(self):
        self.L = int(sqrt(self.hilbert.size))
        assert self.L**2 == self.hilbert.size

        features = _get_feature_list(self)
        self._layers = [
            FastMaskedConv2D(
                L=self.L,
                features=features[i],
                kernel_size=self.kernel_size,
                kernel_dilation=self.kernel_dilation,
                exclusive=(i == 0),
                use_bias=self.use_bias,
                param_dtype=self.param_dtype,
                precision=self.precision,
                kernel_init=self.kernel_init,
                bias_init=self.bias_init,
            )
            for i in range(self.layers)
        ]

    def reshape_inputs(self, inputs: Array) -> Array:
        return inputs.reshape((inputs.shape[0], self.L, self.L))



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import flax as _flax

from .activation import (
    celu,
    elu,
    gelu,
    glu,
    leaky_relu,
    log_sigmoid,
    log_softmax,
    relu,
    sigmoid,
    soft_sign,
    softmax,
    softplus,
    swish,
    silu,
    tanh,
    cosh,
    sinh,
    log_cosh,
    log_sinh,
    log_tanh,
    reim_selu,
    reim_relu,
    reim,
)

from .symmetric_linear import (
    DenseSymm,
    DenseEquivariant,
)

from .masked_linear import MaskedDense1D, MaskedConv1D, MaskedConv2D
from .fast_masked_linear import FastMaskedDense1D, FastMaskedConv1D, FastMaskedConv2D

from .utils import (
    to_array,
    to_matrix,
    split_array_mpi,
    binary_encoding,
)

from . import blocks



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket.utils.moduletools import export, hide_unexported

from .axial import inversion_group as _inv_group
from .axial import C as _C
from ._point_group import PointGroup


hide_unexported(__name__)

__all__ = [
    "icosahedral_rotations",
    "icosahedral",
]


@export
def I() -> PointGroup:  # noqa: E743
    """
    Rotational symmetries of an icosahedron with two vertices on z axis
    and two others in the xz plane.
    """
    g1 = _C(5, axis=(0, 0, 1))
    g2 = _C(5, axis=(2 / 5**0.5, 0, 1 / 5**0.5))
    g = (g2 @ g1).remove_duplicates()
    g = (g @ g2).remove_duplicates()
    g = (g @ g1).remove_duplicates()
    return g


icosahedral_rotations = I


@export
def Ih() -> PointGroup:
    """
    Symmetry group of an icosahedron with two vertices on z axis
    and two others in the xz plane.
    """
    return _inv_group() @ I()


icosahedral = Ih



# Copyright 2023-2024 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from collections.abc import Callable

import numpy as np

import jax
import jax.numpy as jnp

from netket.utils.types import Array
from netket.utils import struct, StaticRange
from netket.utils.dispatch import dispatch

from netket.hilbert.constraint import (
    DiscreteHilbertConstraint,
    ExtraConstraint,
)

from .base import HilbertIndex
from .uniform_tensor import UniformTensorProductHilbertIndex


@dispatch.abstract
def optimalConstrainedHilbertindex(local_states, size, constraint):
    """
    Returns the optimal Hilbert Index to index into the uniform
    state with local degrees of freedom `local_states`, size sites
    and given constraint.

    This function uses dispatch to select a potential optimal implementation,
    and generally returns a default implementation if None better is available.

    Args:
        local_states: The StaticRange Local states.
        size: integer of the number of degrees of freedom.
        constraint: callable class implementing the constraint.
    """


# Generic dispatch rule based on a lookup table. Not very performant, but will always
# work.
@optimalConstrainedHilbertindex.dispatch
def optimalConstrainedHilbertindex_generic(local_states, size, constraint):
    bare_index = UniformTensorProductHilbertIndex(local_states, size)
    return ConstrainedHilbertIndex(bare_index, constraint)


# Rule for ExtraConstraint, which is a constraint that is a sum of two constraints.
# This wraps an optimised hilbert index into a cosntrained one if it exists.
@optimalConstrainedHilbertindex.dispatch
def optimalConstrainedHilbertindex(local_states, size, constraint: ExtraConstraint):
    # If we have a constraint, we tentatively construct a specialised Hilbert index for that particular constraint.
    # If this specialised indexer object exists, we check whether it is more efficient than the generic
    # ConstrainedHilbertIndex one. If it is more efficient, we use it, otherwise we keep the generic one.
    bare_index = optimalConstrainedHilbertindex(
        local_states, size, constraint.base_constraint
    )
    return ConstrainedHilbertIndex(bare_index, constraint.extra_constraint)


@struct.dataclass
class ConstrainedHilbertIndex(HilbertIndex):
    """
    Indexes a constrained hilbert space with a generic constraint function,
    by building an internal lookup table of all states in the constrained space.

    Requires that the unconstrained index is indexable.
    """

    unconstrained_index: HilbertIndex
    constraint_fun: DiscreteHilbertConstraint = struct.field(pytree_node=False)

    @property
    def _bare_numbers(self) -> Array:
        return compute_constrained_to_bare_conversion_table(
            self.unconstrained_index, self.constraint_fun
        )

    @property
    def n_states(self) -> int:
        return self._bare_numbers.shape[0]

    @property
    def size(self) -> int:
        return self.unconstrained_index.size

    @property
    def local_states(self) -> Array | StaticRange:
        return self.unconstrained_index.local_states

    @property
    def local_size(self) -> int:
        return self.unconstrained_index.local_size

    @jax.jit
    def states_to_numbers(self, states: Array) -> Array:
        out = self.unconstrained_index.states_to_numbers(states)
        return jnp.searchsorted(self._bare_numbers, out)

    @jax.jit
    def numbers_to_states(self, numbers: Array) -> Array:
        # convert to original space
        numbers = self._bare_numbers[numbers]
        return self.unconstrained_index.numbers_to_states(numbers)

    def all_states(self) -> Array:
        return self.numbers_to_states(jnp.arange(self.n_states))

    @property
    def is_indexable(self) -> bool:
        return self.unconstrained_index.is_indexable

    @property
    def n_states_bound(self) -> int:
        """
        Returns an integer upper bound on the total number of states.
        Used as a proxy of the computational cost of using this indexer object,
        and by Hilbert spaces to decide which implementation to pick.
        """
        return self.unconstrained_index.n_states


# This function has exponential runtime in self.size, so we cache it in order to
# only compute it once.
# TODO: distribute over devices/MPI (expensive constraint_fun),  choose better chunk size
@partial(jax.jit, static_argnames=("chunk_size", "constraint_fun"))
def compute_constrained_to_bare_conversion_table(
    hilbert_index: HilbertIndex,
    constraint_fun: Callable[[Array], Array],
    *,
    chunk_size: int = 65536,
):
    """
    Computes the conversion table that converts the 'constrained' indices
    of an hilbert space to bare indices, so that routines generating
    only values in an unconstrained space can be used.

    Args:
        hilbert_index:
            A dataclass with only metadata (only pytree_node=False)
        constraint_fun:
            A dataclass with only metadata (only pytree_node=False) and __call__ attribute
            Python functions can be used by wrapping them in a jax.tree_util.Partial
            with no args and keywords.
        chunk_size: (optional, default=65536)
            This function operates on blocks of `chunk_size` states at a time in order
            to lower the memory cost. The default chunk size has been chosen arbitrarily
            and might need tweaking depending on the particular constraint_fun.
    """

    with jax.ensure_compile_time_eval():
        n_chunks = int(np.ceil(hilbert_index.n_states / chunk_size))
        bare_number_chunks = []
        for i in range(n_chunks):
            id_start = chunk_size * i
            id_end = np.minimum(chunk_size * (i + 1), hilbert_index.n_states)
            ids = jnp.arange(id_start, id_end, dtype=jnp.int32)
            states = hilbert_index.numbers_to_states(ids)
            is_constrained = constraint_fun(states)
            (chunk_bare_number,) = jnp.nonzero(is_constrained)
            bare_number_chunks.append(chunk_bare_number + id_start)
        bare_numbers = jnp.concatenate(bare_number_chunks)
    return bare_numbers



# Copyright 2021 The NetKet Authors - All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import flax.linen as nn
import jax.numpy as jnp
from jax.nn.initializers import normal
from flax.linen.dtypes import promote_dtype
from netket.utils.types import DType, Array, NNInitFunc


class Jastrow(nn.Module):
    r"""
    Jastrow wave function :math:`\Psi(s) = \exp(\sum_{i \neq j} s_i W_{ij} s_j)`,
    where W is a symmetric matrix.

    The matrix W is treated as low triangular to avoid
    redundant parameters in the computation.
    """

    param_dtype: DType = jnp.complex128
    """The dtype of the weights."""
    kernel_init: NNInitFunc = normal()
    """Initializer for the weights."""

    @nn.compact
    def __call__(self, x_in: Array):
        nv = x_in.shape[-1]
        il = jnp.tril_indices(nv, k=-1)

        kernel = self.param(
            "kernel", self.kernel_init, (nv * (nv - 1) // 2,), self.param_dtype
        )

        # .at[].set is VERY slow for complex128 numbers in jax.
        # So we do it on the real-valued real and imaginary parts separately and then join them back
        # See issue https://github.com/jax-ml/jax/issues/24872
        if jnp.issubdtype(self.param_dtype, jnp.complex128):
            Wr = (
                jnp.zeros((nv, nv), dtype=kernel.real.dtype)
                .at[il]
                .set(kernel.real, unique_indices=True, indices_are_sorted=True)
            )
            Wi = (
                jnp.zeros((nv, nv), dtype=kernel.imag.dtype)
                .at[il]
                .set(kernel.imag, unique_indices=True, indices_are_sorted=True)
            )
            W = Wr + 1j * Wi

        else:
            W = (
                jnp.zeros((nv, nv), dtype=self.param_dtype)
                .at[il]
                .set(kernel, unique_indices=True, indices_are_sorted=True)
            )

        W, x_in = promote_dtype(W, x_in, dtype=None)
        y = jnp.einsum("...i,ij,...j", x_in, W, x_in)

        return y



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial, wraps

import jax
import jax.numpy as jnp
import numpy as np
from numbers import Number

from jax.tree_util import register_pytree_node_class

from netket.operator import DiscreteJaxOperator
from netket.hilbert.abstract_hilbert import AbstractHilbert
from netket.utils.types import DType

from .base import FermionOperator2ndBase
from .utils import _is_diag_term


@partial(jax.vmap, in_axes=(0, None, None))
def _flip_daggers_split_cast_term_part(term, site_dtype, dagger_dtype):
    # splits sites and daggers out of terms, casts to desired dtype
    sites, daggers = jnp.array(term).reshape([-1, 2]).T
    # we flip the daggers so that the operator returns xp s.t. <xp|O|x> != 0
    daggers = jnp.ones_like(daggers) - daggers
    return sites.astype(site_dtype), daggers.astype(dagger_dtype)


def prepare_terms_list(
    operators,
    site_dtype=np.uint32,
    dagger_dtype=np.int8,
    weight_dtype=jnp.float64,
):
    # return xp s.t. <x|O|xp> != 0
    # see https://github.com/netket/netket/issues/1385
    term_dagger_split_fn = _flip_daggers_split_cast_term_part

    # group the terms together with respect to the number of sites they act on
    terms_dicts = {}
    for t, w in operators.items():
        l = len(t)
        d = terms_dicts.get(l, {})
        d[t] = w
        terms_dicts[l] = d

    res = []
    for d in terms_dicts.values():
        w = jnp.array(list(d.values()), dtype=weight_dtype)
        t = np.array(list(d.keys()), dtype=int)
        res.append((w, *term_dagger_split_fn(t, site_dtype, dagger_dtype)))
    return res


@partial(jax.jit, static_argnums=4)
def _apply_term_scan(x, weight, sites, daggers, unroll=1):
    # here we do jordan wigner:
    # for every site:
    # (1.) destroy/create a particle on current site based the value of dagger
    #      using the raising/lowering operators σ⁺ and σ⁻
    #      where σ⁺|0⟩=|1⟩  and σ⁻|0⟩=0
    #            σ⁺|1⟩=0        σ⁻|1⟩=|0⟩
    # (2.) apply σᶻ to all sites before the current site

    assert daggers.dtype == jnp.bool_
    assert x.dtype == jnp.bool_

    if len(sites) == 0:  # constant diagonal term
        return x, jnp.full(x.shape[:-1], weight), jnp.full(x.shape[:-1], True)

    n_orbitals = x.shape[-1]

    sgn = jnp.zeros(x.shape[:-1], dtype=jnp.bool_)
    zero = jnp.zeros(x.shape[:-1], dtype=jnp.bool_)
    init = x, sgn, zero
    xs = sites, daggers

    def f(carry, xs):
        site, dagger = xs
        x_, sgn, zero = carry

        # apply σ⁻ / σ⁺
        x_new = x_.at[..., site].set(dagger)

        # compute sign from σᶻ (stored as 0/1 for +1/-1)
        mask_all_up_to_site = jnp.arange(n_orbitals, dtype=sites.dtype) < site
        x_masked = x_ & mask_all_up_to_site
        sgn = sgn ^ _reduce_xor(x_masked, (x_.ndim - 1,))

        # check if we did σ⁺|1⟩=0 or σ⁻|0⟩=0
        zero = zero | (x_.at[..., site].get() == dagger)

        return (x_new, sgn, zero), None

    # scan over the sites the term is acting on
    (x_final, sgn, zero), _ = jax.lax.scan(f, init, xs, unroll=unroll)

    # compute the real value of the sign (map [0,1] ↦ [+1,-1])
    sign = 1 - 2 * sgn.astype(weight.dtype)
    # compute the final coefficient
    not_zero = ~zero
    w_final = weight * not_zero * sign
    # return the xp, the mel and wether mel is zero
    return x_final, w_final, not_zero


@partial(jax.vmap, in_axes=(None, 0, 0, 0, None), out_axes=(-2, -1, -1))
def _apply_terms_scan(x, w, sites, daggers, unroll):
    return _apply_term_scan(x, w, sites, daggers, unroll=unroll)


@partial(jax.jit, static_argnums=4)
def apply_terms_scan(x, w, sites, daggers, unroll=1):
    x_res, *res = _apply_terms_scan(
        x.astype(jnp.bool_), w, sites, daggers.astype(jnp.bool_), unroll
    )
    return x_res.astype(x.dtype), *res


def _biti(i, N, dtype=np.uint8):
    bitwidth = 8 * dtype().itemsize
    n, r = divmod(N, bitwidth)
    if r > 0:
        n = n + 1  # padding
    x = jnp.zeros(n, dtype=dtype)

    i, ib = jnp.divmod(i, bitwidth)
    ib = ib.astype(dtype)
    # x is uint there fore we take x.shape[0]-i-1, as -i-1 would underflow
    return x.at[i].set(jax.lax.shift_left(dtype(1), ib)), i


@partial(jnp.vectorize, signature="()->(n)", excluded=(1,))
def biti(i, N, dtype=np.uint8):
    res, _ = _biti(i, N, dtype=dtype)
    return res


@partial(jnp.vectorize, signature="()->(n)", excluded=(1,))
def bituptoi(i, N, dtype=np.uint8):
    mask, byte_index = _biti(i, N, dtype=dtype)
    n = mask.shape[-1]
    return jax.lax.select(jnp.arange(n) <= byte_index, mask - dtype(1), mask)


def _reduce_xor(x, axes):
    return jax.lax.reduce_xor_p.bind(x, axes=tuple(axes))


def _reduce_or(x, axes):
    return jax.lax.reduce_or_p.bind(x, axes=tuple(axes))


def _apply_term_scan_bits(
    x, weight, sites, daggers, unroll=1, process=True, n_orbitals=None
):
    if len(sites) == 0:  # constant diagonal term
        return x, jnp.full(x.shape[:-1], weight), jnp.full(x.shape[:-1], True)

    assert daggers.dtype in [jnp.bool_, jnp.uint8]
    if process:
        n_orbitals = x.shape[-1]
        # x = x.astype(jnp.uint8)
        assert x.dtype == jnp.uint8
        xb = jnp.packbits(x, axis=-1, bitorder="little")
    else:
        assert n_orbitals is not None
        assert x.dtype == jnp.uint8
        xb = x
    # TODO precompute those ?
    # site_mask = biti(sites, n_orbitals)
    # sign_mask = bituptoi(sites, n_orbitals)

    # TODO is unpackbits + xor_reduce faster?
    _parity = lambda x: jax.lax.rem(jax.lax.population_count(x), jnp.uint8(2))

    sgn = jnp.zeros(x.shape[:-1], dtype=jnp.uint8)
    zero = jnp.zeros(x.shape[:-1], dtype=jnp.uint8)

    init = xb, sgn, zero
    xs = sites, daggers

    def f(carry, xs):
        site, dagger = xs
        x_, sgn, zero = carry
        site_mask = biti(site, n_orbitals)
        sign_mask = bituptoi(site, n_orbitals)

        # apply σ⁻ / σ⁺
        # here we (arbitrarily) do
        # σ⁺|1⟩ = 0 |0⟩ and σ⁻|0⟩ = 0 |1⟩
        # as flipping is cheaper than setting/unsetting a bit
        x_new = x_ ^ site_mask

        # compute sign from σᶻ (stored as 0/1 for +1/-1)
        sgn = sgn ^ _reduce_xor(x_ & sign_mask, (x_.ndim - 1,))

        # check if we did σ⁺|1⟩=0 or σ⁻|0⟩=0
        tmp = (x_ ^ ((1 - dagger) * 0xFF)) & site_mask
        zero = zero | _reduce_or(tmp, axes=(tmp.ndim - 1,))

        return (x_new, sgn, zero), None

    (x_final, sgn, zero), _ = jax.lax.scan(f, init, xs, unroll=unroll)
    sign = 1 - 2 * _parity(sgn).astype(weight.dtype)
    not_zero = jax.lax.population_count(zero) == 0
    w_final = weight * not_zero * sign
    # TODO only unpack after we trashed the zeros in get_conn_padded_jax
    # TO avoid extra work
    if process:
        x_final = jnp.unpackbits(x_final, count=n_orbitals, bitorder="little", axis=-1)
    return x_final.astype(x.dtype), w_final, not_zero


@partial(jax.vmap, in_axes=(None, 0, 0, 0, None, None, None), out_axes=(-2, -1, -1))
def _apply_terms_scan_bits(x, w, sites, daggers, unroll, process, n_orbitals):
    return _apply_term_scan_bits(
        x, w, sites, daggers, unroll=unroll, process=process, n_orbitals=n_orbitals
    )


@partial(jax.jit, static_argnums=(4, 5, 6))
def apply_terms_scan_bits(
    x, w, sites, daggers, unroll=1, process=True, n_orbitals=None
):
    x_res, *res = _apply_terms_scan_bits(
        x.astype(jnp.uint8),
        w,
        sites,
        daggers.astype(jnp.uint8),
        unroll,
        process,
        n_orbitals,
    )
    return x_res.astype(x.dtype), *res


# mostly masks, some indexing
def _apply_term_masks(x, w, sites, daggers):
    # sites can be an unsigned int
    # daggers and x need to be signed, preferably of the same type
    # assert x.dtype == np.int8
    # assert daggers.dtype == np.int8

    if len(sites) == 0:  # constant diagonal term
        return x, jnp.full(x.shape[:-1], w), jnp.full(x.shape[:-1], True)

    no = x.shape[-1]
    fill_vec = jnp.arange(no, dtype=sites.dtype)
    masks_flip = jnp.eye(no, dtype=x.dtype)[sites]
    masks_sgn = (fill_vec[None] < sites[:, None]).astype(x.dtype)
    daggers_pm = (daggers - (1 - daggers)).astype(x.dtype)
    add_flip = masks_flip * daggers_pm[:, None]
    add_flip_padded = jnp.vstack([jnp.zeros_like(add_flip[..., 0, :]), add_flip])
    add_flip_cum = jnp.cumsum(add_flip_padded, axis=-2)
    x_at_i = x[..., None, :] + add_flip_cum[(None,) * (x.ndim - 1)]
    x_final_ = x_at_i[..., -1, :]
    x_at_i = x_at_i[..., :-1, :]
    x_final = jnp.clip(x_final_, 0, 1)
    r = jnp.remainder(jnp.einsum("...ij,ij -> ...", x_at_i, masks_sgn), 2)
    sgn = -1 * r + (1 - r)
    d = x_at_i != daggers[:, None]
    xi = x_at_i[..., jnp.arange(len(sites), dtype=np.uint32), sites]
    d = xi != daggers
    # here we cast, for the case when x is float64 but weights are float32
    # and jax would promote the result to float64
    w_final = w * (d.prod(axis=-1) * sgn).astype(w.dtype)
    return x_final, w_final, d.prod(axis=-1)


@partial(jax.vmap, in_axes=(None, 0, 0, 0), out_axes=(-2, -1, -1))
def apply_terms_masks(x, w, sites, daggers):
    x_res, *res = _apply_term_masks(
        x.astype(np.int8), w, sites, daggers.astype(np.int8)
    )
    return x_res.astype(x.dtype), *res


# only masks
def _apply_term_only_masks(x, w, sites, daggers):
    # sites can be an unsigned int
    # daggers and x need to be signed, preferably of the same type

    if not jnp.issubdtype(x.dtype, jnp.signedinteger):
        if jnp.issubdtype(x.dtype, jnp.floating):
            pass  # allow float for the time being
        else:
            raise ValueError(
                f"x has incompatible type. expect a signed integer but got {x.dtype}"
            )
    if not jnp.issubdtype(daggers.dtype, jnp.signedinteger):
        raise ValueError(
            f"daggers has incompatible type. expect a signed integer but got {daggers.dtype}"
        )
    if not jnp.issubdtype(sites.dtype, jnp.integer):
        raise ValueError(
            f"sites has incompatible type. expect a integer but got {sites.dtype}"
        )

    # for daggers it's crucial its a signed int, we need it to go negative
    # (we might be able to get away using underflow if we are careful not to cast,
    # but let's not rely on it)

    if len(sites) == 0:  # constant diagonal term
        return x, jnp.full(x.shape[:-1], w), jnp.full(x.shape[:-1], True)

    # TODO precomute all those masks, and pass them instead of the index?

    n_orbitals = x.shape[-1]
    # mask for the sites we are acting on
    # ensure it's the same dtype as daggers
    # e.g. [0,0,1,0]
    ara = jnp.arange(n_orbitals, dtype=sites.dtype)
    masks_site = ara[None] == sites[:, None]

    # mask for all sites up to (not including) each site to flip (will be needed for the jordan-wigner)
    # e.g. [1,1,0,0]
    masks_all_up_to_site = ara < sites[:, None]

    # here we do jordan wigner
    # 1. we start with the raising / lowering operators, as given by daggers
    # raising creates a particle, so we have to add +1
    # and lowering destroys one, so we have to add -1

    # map [0,1] -> [-1, +1] ( by taking 2x-1)
    # be careful about type as daggers_pm can and will be negative
    daggers_pm = daggers - (1 - daggers)
    # compute the action of each creation/annihilation operator in the term
    add_flip = masks_site * daggers_pm[:, None]

    # add a first row of 0, for starting the cumsum
    add_flip_padded = jnp.concatenate(
        [jnp.zeros_like(add_flip[..., :1, :]), add_flip], axis=-2
    )
    # now compute the cumulative action, i.e. what was applied to the state when operator i sees it
    # the first operator gets the initial state which is why we just padded with 0 meaning do nothing
    add_flip_cum = jnp.cumsum(add_flip_padded, axis=-2)

    # now apply the actions we just computed
    # this gives us the state when operator i sees it, after all up to i have been applied
    x_at_i = x[..., None, :] + add_flip_cum[(None,) * (x.ndim - 1)]

    # the last one (remember, we padded) is the final state:
    x_final_ = x_at_i[..., -1, :]
    # remove final state, now its the state when operator i sees it
    x_at_i = x_at_i[..., :-1, :]

    # we might have just tried to create a fermion in an already
    # occupied orbital, or destroyed one in an empty orbital
    # the resulting matrix element will be zero
    # to get a valid state here we clip, although we could also just set it to the original state
    x_final = jnp.clip(x_final_, 0, 1)

    # compute where we created a fermion in an already occupied orbital or destroyed one in an empty orbital
    # both will result in mel 0
    # version with indexing
    # xi = x_at_i[..., jnp.arange(len(sites), dtype=np.uint32), sites]
    # not_illegal = (xi != daggers[None]).all(axis=-1)
    # version with masks
    illegal = ((x_at_i == daggers[:, None]) * masks_site).any(axis=(-2, -1))
    not_illegal = ~illegal

    # 2. now do the Z gates for the jordan-wigner
    # (for an operator on site i apply Z on all up to site i)

    # jordan-wigner sign
    sgn = 1 - 2 * jnp.remainder(
        jnp.einsum("...ij,ij -> ...", x_at_i, masks_all_up_to_site), 2
    )

    # compute the final matrix element
    # here we cast, for the case when x is float64 but weights are float32
    # and jax would promote the result to float64
    w_final = w * (not_illegal * sgn).astype(w.dtype)
    # we assume w is not 0, otherwise we would have already removed it
    return x_final, w_final, not_illegal


@partial(jax.vmap, in_axes=(None, 0, 0, 0), out_axes=(-2, -1, -1))
def apply_terms_only_masks(x, w, sites, daggers):
    # force cast to int8
    # x_res, *res = _apply_term_only_masks(x.astype(np.int8), w, sites, daggers.astype(np.int8))
    # return x_res.astype(x.dtype), *res

    # run calculations in type of x
    # TODO make sure we have the correct type a priori, and avoid casting here
    daggers = daggers.astype(np.int8)
    return _apply_term_only_masks(x, w, sites, daggers)


# default to unroll=4, which means for chemistry we unroll everything
# seems faster on gpu
@partial(jax.jit, static_argnums=(0, 1, 5))
def get_conn_padded_jax(
    max_conn_size,
    dtype,
    tl_diag,
    tl_offdiag,
    x,
    apply_terms_fun=partial(apply_terms_scan, unroll=4),
):
    # dtype arg is only needed for the empty case when there are no terms

    if len(tl_diag) == 0 and len(tl_offdiag) == 0:
        xp = x[..., None, :][..., :0, :]
        mels = jnp.zeros(xp.shape[:-1], dtype=dtype)
        n_conn = np.zeros(mels.shape, dtype=int)
        return xp, mels, n_conn

    if len(tl_diag) > 0:
        weight_dtype = tl_diag[-1][0].dtype
        assert weight_dtype == dtype
    if len(tl_offdiag) > 0:
        weight_dtype = tl_offdiag[-1][0].dtype
        assert weight_dtype == dtype

    xp_list = []
    mels_list = []
    nonzero_mask_list = []

    # all terms in the diagonal have the same final state,
    # we sum the mels
    xp_diag_ = x[..., None, :]
    mel_diag_ = jnp.zeros(xp_diag_.shape[:-1], dtype=weight_dtype)
    nonzero_mask_ = jnp.ones(mel_diag_.shape, dtype=jnp.bool_)
    if len(tl_diag) == 0:
        xp_diag_ = xp_diag_[..., :0, :]
        mel_diag_ = mel_diag_[..., :0]
        nonzero_mask_ = nonzero_mask_[..., :0]
    else:
        # iterate over the different length terms (0, 2, 4, ...)
        for w, sites, daggers in tl_diag:
            # we trash xp, dce will make sure we don't even compute it
            _, mels_, _ = apply_terms_fun(x, w, sites, daggers)
            mel_diag_ = mel_diag_ + mels_.sum(axis=-1, keepdims=True)
    # TODO here we could check if the diagonal is < cutoff and set nonzero_mask_ to False

    xp_list.append(xp_diag_)
    mels_list.append(mel_diag_)
    nonzero_mask_list.append(nonzero_mask_)
    # iterate over the different length terms (0, 2, 4, ...)
    for w, sites, daggers in tl_offdiag:
        xp_, mels_, nonzero_mask_ = apply_terms_fun(x, w, sites, daggers)
        xp_list.append(xp_)
        mels_list.append(mels_)
        nonzero_mask_list.append(nonzero_mask_)

    # pad with 0 and old state
    xp_list.append(x[..., None, :])
    mels_list.append(jnp.zeros((x.shape[:-1] + (1,)), dtype=weight_dtype))

    xp_padded = jnp.concatenate(xp_list, axis=-2)
    mels_padded = jnp.concatenate(mels_list, axis=-1)
    nonzero_mask = jnp.concatenate(nonzero_mask_list, axis=-1)

    # move the nonzeros to the beginning

    n_nonzero = nonzero_mask.sum(axis=-1)
    _nonzero_fn = partial(jnp.where, size=max_conn_size, fill_value=-1)
    (i_nonzero,) = jnp.vectorize(_nonzero_fn, signature="(i)->(j)")(nonzero_mask)
    xp_u = jnp.take_along_axis(xp_padded, i_nonzero[..., None], axis=-2)
    mels_u = jnp.take_along_axis(mels_padded, i_nonzero, axis=-1)

    # TODO here would be the place to remove / merge repeated mels
    #
    # you should check that n_nonzero <= max_conn_size outside of jit,
    # and increase max_conn_size if it's not
    return xp_u, mels_u, n_nonzero


@partial(jax.jit, static_argnums=(0, 4))
def n_conn_jax(dtype, tl_diag, tl_offdiag, x, apply_terms_fun=apply_terms_scan):
    max_conn_size = 0
    # let dce take care of not computing xp
    _, _, n_conn = get_conn_padded_jax(
        max_conn_size,
        dtype,
        tl_diag,
        tl_offdiag,
        x,
        apply_terms_fun=partial(apply_terms_scan, unroll=4),
    )
    return n_conn


@register_pytree_node_class
class FermionOperator2ndJax(FermionOperator2ndBase, DiscreteJaxOperator):
    r"""
    A fermionic operator in :math:`2^{nd}` quantization using pure jax dark
    magic for indexing.

    .. warning::

        This class is a Pytree, so it **can** be used inside of jax-transformed
        functions like `jax.grad` or `jax.jit`.

    When using native (experimental) sharding, or when working with GPUs,
    we reccomend using this implementations of the operators for potentially
    better performance.

    Can be converted to the Numba one with the method
    :meth:`~netket.operator.FermionOperator2ndJax.to_numba_operator()`.
    """

    @wraps(FermionOperator2ndBase.__init__)
    def __init__(
        self,
        hilbert: AbstractHilbert,
        terms: list[str] | list[list[list[int]]] = None,
        weights: list[float | complex] | None = None,
        constant: Number = 0,
        cutoff: float = 1e-10,
        dtype: DType = None,
        _mode: str = "scan",
    ):
        super().__init__(
            hilbert, terms, weights, constant=constant, cutoff=cutoff, dtype=dtype
        )
        self._mode = _mode

    @property
    def _mode(self):
        """
        (Internal) Indexing mode of the operator.

        Valid values are "index" or "mask".

        'Index' uses the standard LocalOperator-like indexing of changed points,
        while the latter uses constant-size masks.

        The latter does not really need recompilation for paulistrings with
        different values, and this could be changed in the future.
        """
        return self._mode_attr

    @_mode.setter
    def _mode(self, mode):
        avaible_modes = ["scan", "mask"]

        if mode not in avaible_modes:
            raise ValueError(
                f"unknown mode {mode}. Available modes are {avaible_modes}."
            )
        self._mode_attr = mode

    def _setup(self, force: bool = False):
        """Analyze the operator strings and precompute arrays for get_conn inference"""
        if force or not self._initialized:
            super()._setup(self)
            # TODO ideally we would set dagger_dtype to the same as x
            # however, unfortunately, the dtype of the states in netket
            # is stored in the sampler and not in hilbert, so we don't know it at this stage
            diag_operators = {
                k: v for k, v in self._operators.items() if _is_diag_term(k)
            }
            offdiag_operators = {
                k: v for k, v in self._operators.items() if not _is_diag_term(k)
            }

            self._terms_list_diag = prepare_terms_list(
                diag_operators,
                site_dtype=np.uint32,
                dagger_dtype=jnp.bool_,
                weight_dtype=self._dtype,
            )
            self._terms_list_offdiag = prepare_terms_list(
                offdiag_operators,
                site_dtype=np.uint32,
                dagger_dtype=jnp.bool_,
                weight_dtype=self._dtype,
            )

            # TODO the following could be reduced further
            self._max_conn_size = int(len(self._terms_list_diag) > 0) + len(
                offdiag_operators
            )
            self._initialized = True

    def tree_flatten(self):
        self._setup()
        data = (
            self._terms_list_diag,
            self._terms_list_offdiag,
        )
        metadata = {
            "hilbert": self.hilbert,
            "operators": self._operators,
            "dtype": self.dtype,
            "max_conn_size": self._max_conn_size,
        }
        return data, metadata

    @classmethod
    def tree_unflatten(cls, metadata, data):
        hi = metadata["hilbert"]
        dtype = metadata["dtype"]

        op = cls(hi, [], [], dtype=dtype)

        op._operators = metadata["operators"]
        op._max_conn_size = metadata["max_conn_size"]
        op._initialized = True
        (op._terms_list_diag, op._terms_list_offdiag) = data
        return op

    def to_numba_operator(self) -> "FermionOperator2nd":  # noqa: F821
        """
        Returns the standard numba version of this operator, which is an
        instance of :class:`netket.operator.FermionOperator2nd`.
        """
        from .numba import FermionOperator2nd

        new_op = FermionOperator2nd(self.hilbert, cutoff=self._cutoff, dtype=self.dtype)
        new_op._operators = self._operators.copy()
        return new_op

    def get_conn_padded(self, x):
        self._setup()

        if self._mode == "scan":
            apply_terms_fun = apply_terms_scan_bits
        elif self._mode == "mask":
            apply_terms_fun = apply_terms_masks

        xp, mels, _ = get_conn_padded_jax(
            self._max_conn_size,
            self._dtype,
            self._terms_list_diag,
            self._terms_list_offdiag,
            x,
            apply_terms_fun=apply_terms_fun,
        )
        # TODO if we are outside jit (i don't know how to detect it)
        # we coule check here that _max_conn_size was not too small
        #
        # success = jax.jit(jnp.max)(n_conn) <= self._max_conn_size # jit for gda
        # if not success:
        #     raise ValueError("more connected elements than _max_conn_size")
        #
        # alternatively we could return success
        return xp, mels

    def n_conn(self, x):
        self._setup()
        if self._mode == "scan":
            apply_terms_fun = apply_terms_scan_bits
        elif self._mode == "mask":
            apply_terms_fun = apply_terms_masks

        return n_conn_jax(
            self._dtype,
            self._terms_list_diag,
            self._terms_list_offdiag,
            x,
            apply_terms_fun=apply_terms_fun,
        )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import permutations
from functools import partial
from collections.abc import Sequence
import numpy as np
import warnings

from .lattice import Lattice

from netket.utils.group import PointGroup, PGSymmetry, planar, cubic, Identity
from netket.errors import InitializePeriodicLatticeOnSmallLatticeWarning


def _perm_symm(perm: tuple) -> PGSymmetry:
    n = len(perm)
    M = np.zeros((n, n))
    M[range(n), perm] = 1
    return PGSymmetry(M)


def _axis_reflection(axis: int, ndim: int) -> PGSymmetry:
    M = np.eye(ndim)
    M[axis, axis] = -1
    return PGSymmetry(M)


def _grid_point_group(
    extent: Sequence[int], pbc: Sequence[bool], color_edges: bool
) -> PointGroup:
    """Point group of `Grid`, made up of axis permutations and flipping each axis."""
    ndim = len(extent)
    # Cannot exchange two axes if they are colored differently; otherwise,
    # can only exchange them if they have the same kind of BC and length.
    # Represent open BC by setting kind[i] = -extent[i], so just have to match these
    if color_edges:
        result = PointGroup([Identity()], ndim=ndim)
    else:
        axis_perm = []
        axes = np.arange(ndim, dtype=int)
        extent = np.asarray(extent, dtype=int)
        kind = np.where(pbc, extent, -extent)
        for perm in permutations(axes):
            if np.all(kind == kind[list(perm)]):
                if np.all(perm == axes):
                    axis_perm.append(Identity())
                else:
                    axis_perm.append(_perm_symm(perm))
        result = PointGroup(axis_perm, ndim=ndim)

    # reflections across axes and setting the origin
    # OBC axes are only symmetric w.r.t. their midpoint, (extent[i]-1)/2
    origin = []
    for i in range(ndim):
        result = result @ PointGroup([Identity(), _axis_reflection(i, ndim)], ndim=ndim)
        origin.append(0 if pbc[i] else (extent[i] - 1) / 2)

    return result.change_origin(origin)


def Grid(
    extent: Sequence[int],
    *,
    pbc: bool | Sequence[bool] = True,
    color_edges: bool = False,
    point_group: PointGroup | None = None,
    **kwargs,
) -> Lattice:
    """
    Constructs a hypercubic lattice given its extent in all dimensions.

    Args:
        extent: Size of the lattice along each dimension. It must be a list with
                integer components >= 1.
        pbc: If `True`, the grid will have periodic boundary conditions (PBC);
             if `False`, the grid will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        color_edges: generates nearest-neighbour edges colored according to direction
                     i.e. edges along Cartesian direction #i have color i
                     cannot be used with `max_neighbor_order` or `custom_edges`
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full hypercube symmetry group.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.

    Examples:
        Construct a 5x10 square lattice with periodic boundary conditions:

        >>> import netket
        >>> g=netket.graph.Grid(extent=[5, 10], pbc=True)
        >>> print(g.n_nodes)
        50

        Construct a 2x2x3 cubic lattice with open boundary conditions:

        >>> g=netket.graph.Grid(extent=[2,2,3], pbc=False)
        >>> print(g.n_nodes)
        12
    """
    extent = np.asarray(extent, dtype=int)
    ndim = len(extent)
    if isinstance(pbc, bool):
        pbc = [pbc] * ndim
    raised_periodic_lattice_on_small_lattice_warning = False
    for i in range(ndim):
        if (
            extent[i] <= 2
            and pbc[i]
            and not raised_periodic_lattice_on_small_lattice_warning
        ):
            raised_periodic_lattice_on_small_lattice_warning = True
            warnings.warn(
                InitializePeriodicLatticeOnSmallLatticeWarning(extent[i], i),
                UserWarning,
            )
    if color_edges:
        kwargs["custom_edges"] = [(0, 0, vec) for vec in np.eye(ndim)]
    if point_group is None:
        point_group = partial(_grid_point_group, extent, pbc, color_edges)
    return Lattice(
        basis_vectors=np.eye(ndim),
        extent=extent,
        pbc=pbc,
        point_group=point_group,
        **kwargs,
    )


def Hypercube(length: int, n_dim: int = 1, *, pbc: bool = True, **kwargs) -> Lattice:
    r"""Constructs a hypercubic lattice with equal side length in all dimensions.
    Periodic boundary conditions can also be imposed.

    Args:
        length: Side length of the hypercube; must always be >=1
        n_dim: Dimension of the hypercube; must be at least 1.
        pbc: Whether the hypercube should have periodic boundary conditions
            (in all directions)
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.

    Examples:
         A 10x10x10 cubic lattice with periodic boundary conditions can be
         constructed as follows:

         >>> import netket
         >>> g = netket.graph.Hypercube(10, n_dim=3, pbc=True)
         >>> print(g.n_nodes)
         1000
    """
    if not isinstance(length, int) or length <= 0:
        raise TypeError("Argument `length` must be a positive integer")
    length_vector = [length] * n_dim
    return Grid(length_vector, pbc=pbc, **kwargs)


def Cube(length: int, *, pbc: bool = True, **kwargs) -> Lattice:
    """Constructs a cubic lattice of side `length`
    Periodic boundary conditions can also be imposed

    Args:
        length: Side length of the cube; must always be >=1
        pbc: Whether the cube should have periodic boundary conditions
            (in all directions)
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.

    Examples:
        A 10×10×10 cubic lattice with periodic boundary conditions can be
        constructed as follows:

        >>> import netket
        >>> g=netket.graph.Cube(10, pbc=True)
        >>> print(g.n_nodes)
        1000
    """
    return Hypercube(length, pbc=pbc, n_dim=3, **kwargs)


def Square(length: int, *, pbc: bool = True, **kwargs) -> Lattice:
    """Constructs a square lattice of side `length`
    Periodic boundary conditions can also be imposed

    Args:
        length: Side length of the square; must always be >=1
        pbc: Whether the square should have periodic boundary
            conditions (in both directions)
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.

    Examples:
        A 10x10 square lattice with periodic boundary conditions can be
        constructed as follows:

        >>> import netket
        >>> g=netket.graph.Square(10, pbc=True)
        >>> print(g.n_nodes)
        100
    """
    return Hypercube(length, pbc=pbc, n_dim=2, **kwargs)


def Chain(length: int, *, pbc: bool = True, **kwargs) -> Lattice:
    r"""Constructs a chain of `length` sites.
    Periodic boundary conditions can also be imposed

    Args:
        length: Length of the chain. It must always be >=1
        pbc: Whether the chain should have periodic boundary conditions
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.

    Examples:
        A 10 site chain with periodic boundary conditions can be
        constructed as follows:

        >>> import netket
        >>> g = netket.graph.Chain(10, pbc=True)
        >>> print(g.n_nodes)
        10
    """
    return Hypercube(length, pbc=pbc, n_dim=1, **kwargs)


def BCC(
    extent: Sequence[int],
    *,
    pbc: bool | Sequence[bool] = True,
    point_group: PointGroup | None = None,
    **kwargs,
) -> Lattice:
    """Constructs a BCC lattice of a given spatial extent.
    Periodic boundary conditions can also be imposed
    Sites are returned at the Bravais lattice points.

    Arguments:
        extent: Number of primitive unit cells along each direction, needs to be
            an array of length 3
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
             if `False`, the lattice will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full cubic symmetry group.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct a BCC lattice with 3×3×3 primitive unit cells:

        >>> from netket.graph import BCC
        >>> g = BCC(extent=[3,3,3])
        >>> print(g.n_nodes)
        27
    """
    basis = [[-0.5, 0.5, 0.5], [0.5, -0.5, 0.5], [0.5, 0.5, -0.5]]
    if point_group is None:
        # determine if full point group is realised by the simulation box
        point_group = cubic.Oh() if np.all(pbc) and len(set(extent)) == 1 else None
    return Lattice(
        basis_vectors=basis, extent=extent, pbc=pbc, point_group=point_group, **kwargs
    )


def FCC(
    extent: Sequence[int],
    *,
    pbc: bool | Sequence[bool] = True,
    point_group: PointGroup | None = None,
    **kwargs,
) -> Lattice:
    """Constructs an FCC lattice of a given spatial extent.
    Periodic boundary conditions can also be imposed
    Sites are returned at the Bravais lattice points.

    Arguments:
        extent: Number of primitive unit cells along each direction, needs
            to be an array of length 3
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
             if `False`, the lattice will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full cubic symmetry group.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct an FCC lattice with 3×3×3 primitive unit cells:

        >>> from netket.graph import FCC
        >>> g = FCC(extent=[3,3,3])
        >>> print(g.n_nodes)
        27
    """
    basis = [[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]]
    if point_group is None:
        # determine if full point group is realised by the simulation box
        point_group = cubic.Oh() if np.all(pbc) and len(set(extent)) == 1 else None
    return Lattice(
        basis_vectors=basis, extent=extent, pbc=pbc, point_group=point_group, **kwargs
    )


def Diamond(
    extent: Sequence[int],
    *,
    pbc: bool | Sequence[bool] = True,
    point_group: PointGroup | None = None,
    **kwargs,
) -> Lattice:
    """Constructs a diamond lattice of a given spatial extent.
    Periodic boundary conditions can also be imposed.

    Sites are returned at the 8a Wyckoff positions of the FCC lattice
    ([000], [1/4,1/4,1/4], and translations thereof).

    Arguments:
        extent: Number of primitive unit cells along each direction, needs to
            be an array of length 3
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
             if `False`, the lattice will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full symmetry group of the diamond lattice.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct a diamond lattice with 3×3×3 primitive unit cells:

        >>> from netket.graph import Diamond
        >>> g = Diamond(extent=[3,3,3])
        >>> print(g.n_nodes)
        54
    """
    basis = [[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]]
    sites = [[0, 0, 0], [0.25, 0.25, 0.25]]
    if point_group is None:
        # determine if full point group is realised by the simulation box
        point_group = cubic.Fd3m() if np.all(pbc) and len(set(extent)) == 1 else None
    return Lattice(
        basis_vectors=basis,
        site_offsets=sites,
        extent=extent,
        pbc=pbc,
        point_group=point_group,
        **kwargs,
    )


def Pyrochlore(
    extent: Sequence[int],
    *,
    pbc: bool | Sequence[bool] = True,
    point_group: PointGroup | None = None,
    **kwargs,
) -> Lattice:
    """Constructs a pyrochlore lattice of a given spatial extent.
    Periodic boundary conditions can also be imposed.

    Sites are returned at the 16c Wyckoff positions of the FCC lattice
    ([111]/8, [1 -1 -1]/8, [-1 1 -1]/8, [-1 -1 1]/8, and translations thereof).

    Arguments:
        extent: Number of primitive unit cells along each direction, needs to be
            an array of length 3
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
            if `False`, the lattice will have open boundary conditions (OBC).
            This parameter can also be a list of booleans with same length as
            the parameter `length`, in which case each dimension will have
            PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full symmetry group of the pyrochlore lattice.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct a pyrochlore lattice with 3×3×3 primitive unit cells:

        >>> from netket.graph import Pyrochlore
        >>> g = Pyrochlore(extent=[3,3,3])
        >>> print(g.n_nodes)
        108
    """
    basis = [[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]]
    sites = np.array([[1, 1, 1], [1, 3, 3], [3, 1, 3], [3, 3, 1]]) / 8
    if point_group is None:
        # determine if full point group is realised by the simulation box
        point_group = cubic.Fd3m() if np.all(pbc) and len(set(extent)) == 1 else None
    return Lattice(
        basis_vectors=basis,
        site_offsets=sites,
        extent=extent,
        pbc=pbc,
        point_group=point_group,
        **kwargs,
    )


def _hexagonal_general(
    extent,
    *,
    site_offsets=None,
    pbc: bool | Sequence[bool] = True,
    point_group: PointGroup | None = None,
    **kwargs,
) -> Lattice:
    basis = [[1, 0], [0.5, 0.75**0.5]]
    if point_group is None:
        # determine if full point group is realised by the simulation box
        point_group = planar.D(6) if np.all(pbc) and extent[0] == extent[1] else None
    return Lattice(
        basis_vectors=basis,
        extent=extent,
        site_offsets=site_offsets,
        pbc=pbc,
        point_group=point_group,
        **kwargs,
    )


def Triangular(extent, *, pbc: bool | Sequence[bool] = True, **kwargs) -> Lattice:
    r"""Constructs a triangular lattice of a given spatial extent.
    Periodic boundary conditions can also be imposed
    Sites are returned at the Bravais lattice points.

    Arguments:
        extent: Number of unit cells along each direction, needs to be an array
            of length 2
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
             if `False`, the lattice will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full hexagonal symmetry group.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct a triangular lattice with 3 × 3 unit cells:

        >>> from netket.graph import Triangular
        >>> g = Triangular(extent=[3, 3])
        >>> print(g.n_nodes)
        9
    """
    ndim = len(extent)
    raised_periodic_lattice_on_small_lattice_warning = False
    if isinstance(pbc, bool):
        pbc = [pbc] * ndim
    for i in range(ndim):
        if (
            extent[i] <= 2
            and pbc[i]
            and not raised_periodic_lattice_on_small_lattice_warning
        ):
            raised_periodic_lattice_on_small_lattice_warning = True
            warnings.warn(
                InitializePeriodicLatticeOnSmallLatticeWarning(extent[i], i),
                UserWarning,
            )
    return _hexagonal_general(extent, site_offsets=None, pbc=pbc, **kwargs)


def Honeycomb(extent, *, pbc: bool | Sequence[bool] = True, **kwargs) -> Lattice:
    r"""Constructs a honeycomb lattice of a given spatial extent.
    Periodic boundary conditions can also be imposed.
    Sites are returned at the 2b Wyckoff positions.

    Arguments:
        extent: Number of unit cells along each direction, needs to be an array
            of length 2
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
             if `False`, the lattice will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full hexagonal symmetry group.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct a honeycomb lattice with 3 × 3 unit cells:

        >>> from netket.graph import Honeycomb
        >>> g = Honeycomb(extent=[3, 3])
        >>> print(g.n_nodes)
        18
    """
    return _hexagonal_general(
        extent,
        site_offsets=[[0.5, 0.5 / 3**0.5], [1, 1 / 3**0.5]],
        pbc=pbc,
        **kwargs,
    )


def Kagome(extent, *, pbc: bool | Sequence[bool] = True, **kwargs) -> Lattice:
    r"""Constructs a kagome lattice of a given spatial extent.
    Periodic boundary conditions can also be imposed.
    Sites are returned at the 3c Wyckoff positions.

    Arguments:
        extent: Number of unit cells along each direction, needs to be an array
            of length 2
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
             if `False`, the lattice will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full hexagonal symmetry group.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct a kagome lattice with 3 × 3 unit cells:

        >>> from netket.graph import Kagome
        >>> g = Kagome(extent=[3, 3])
        >>> print(g.n_nodes)
        27
    """
    return _hexagonal_general(
        extent,
        site_offsets=[[0.5, 0], [0.25, 0.75**0.5 / 2], [0.75, 0.75**0.5 / 2]],
        pbc=pbc,
        **kwargs,
    )


def KitaevHoneycomb(
    extent,
    *,
    pbc: bool | Sequence[bool] = True,
    point_group: PointGroup | None = None,
    **kwargs,
) -> Lattice:
    r"""Constructs a honeycomb lattice of a given spatial extent.

    Nearest-neighbour edges are coloured according to direction
    (cf. Kitaev, https://doi.org/10.1016/j.aop.2005.10.005).
    Periodic boundary conditions can also be imposed.
    Sites are returned at the 2b Wyckoff positions.

    Arguments:
        extent: Number of unit cells along each direction, needs to be an array
            of length 2
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
             if `False`, the lattice will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the 180° rotation symmetry of the Kitaev model.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct a Kitaev honeycomb lattice with 3 × 3 unit cells:

        >>> from netket.graph import KitaevHoneycomb
        >>> g = KitaevHoneycomb(extent=[3, 3])
        >>> print(g.n_nodes)
        18
        >>> print(len(g.edges(filter_color=2)))
        9
    """
    if point_group is None:
        point_group = planar.C(2) if np.all(pbc) else None
    return Lattice(
        basis_vectors=[[1, 0], [0.5, 0.75**0.5]],
        extent=extent,
        site_offsets=[[0.5, 0.5 / 3**0.5], [1, 1 / 3**0.5]],
        pbc=pbc,
        point_group=point_group,
        custom_edges=[
            (0, 1, [0.5, 0.5 / 3**0.5]),
            (0, 1, [-0.5, 0.5 / 3**0.5]),
            (0, 1, [0, -1 / 3**0.5]),
        ],
        **kwargs,
    )
